# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import absolute_import, division, print_function, unicode_literals
import pytest
import numpy as np
import astropy.units as u
from astropy.table import Table
from ...utils.testing import assert_quantity_allclose
from ...utils.testing import requires_dependency, requires_data, mpl_plot_check
from ...utils.energy import EnergyBounds
from ..models import PowerLaw, ConstantModel
from .. import SpectrumObservation, SpectrumFitResult, FluxPoints, SpectrumResult


@pytest.fixture(scope="session")
def fit_result():
    filename = "$GAMMAPY_DATA/joint-crab/spectra/hess/pha_obs23592.fits"
    obs = SpectrumObservation.read(filename)
    best_fit_model = PowerLaw(
        index=2.1, amplitude=1e-11 * u.Unit("cm-2 s-1 TeV-1"), reference=1 * u.TeV
    )
    npred = obs.predicted_counts(best_fit_model).data.data.value
    covariance = np.diag([0.1 ** 2, 1e-12 ** 2, 0])
    best_fit_model.parameters.covariance = covariance
    fit_range = [0.1, 50] * u.TeV
    return SpectrumFitResult(
        model=best_fit_model,
        fit_range=fit_range,
        statname="wstat",
        statval=42,
        npred=npred,
        obs=obs,
    )


@pytest.fixture(scope="session")
def flux_points():
    energies = EnergyBounds.equal_log_spacing(1, 10, 3, unit="TeV")
    dnde = [1e-11, 2e-11, 0.5e-11, 6e-11] * u.Unit("cm-2 s-1 TeV-1")
    dnde_err = np.ones(len(energies)) * 1e-12 * u.Unit("cm-2 s-1 TeV-1")
    dnde_ul = [np.nan, np.nan, np.nan, 6e-11] * u.Unit("cm-2 s-1 TeV-1")
    is_ul = [False, False, False, True]
    table = Table(
        data=[energies, dnde, dnde_err, is_ul, dnde_ul],
        names=["e_ref", "dnde", "dnde_err", "is_ul", "dnde_ul"],
        meta={"SED_TYPE": "dnde"},
    )
    return FluxPoints(table)


@pytest.fixture(scope="session")
def spectrum_result(flux_points):
    model = ConstantModel(const=1e-11 * u.Unit("cm-2 s-1 TeV-1"))
    model.parameters.set_error(0, 1e-10 * u.Unit("cm-2 s-1 TeV-1"))
    return SpectrumResult(model=model, points=flux_points)


@requires_data("gammapy-data")
class TestSpectrumFitResult:
    @requires_dependency("uncertainties")
    def test_basic(self, fit_result):
        assert "PowerLaw" in str(fit_result)
        assert "index" in fit_result.to_table().colnames

    def test_io(self, tmpdir, fit_result):
        filename = tmpdir / "test.yaml"
        fit_result.to_yaml(filename)
        read_result = SpectrumFitResult.from_yaml(filename)
        test_e = 12.5 * u.TeV
        assert_quantity_allclose(fit_result.model(test_e), read_result.model(test_e))

    @requires_dependency("matplotlib")
    def test_plot(self, fit_result):
        with mpl_plot_check():
            fit_result.plot()


@requires_data("gammapy-data")
class TestSpectrumResult:
    def test_basic(self, spectrum_result):
        assert "SpectrumResult" in str(spectrum_result)

    def test_residuals(self, spectrum_result):
        res, res_err = spectrum_result.flux_point_residuals
        assert_quantity_allclose(res, [0, 1, -0.5, np.nan])
        assert_quantity_allclose(res_err, [0.1, 0.1, 0.1, np.nan])

    @requires_dependency("matplotlib")
    @requires_dependency("uncertainties")
    def test_plot(self, spectrum_result):
        with mpl_plot_check():
            spectrum_result.plot(energy_range=[1, 10] * u.TeV, energy_power=2)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Time analysis."""
from .simulate import *
from .lightcurve import *
from .exptest import *
from .period import *
from .plot_periodogram import *



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import absolute_import, division, print_function, unicode_literals
from numpy.testing import assert_allclose
from astropy.units import Quantity
from ...utils.testing import assert_quantity_allclose
from ..powerlaw import (
    power_law_pivot_energy,
    power_law_energy_flux,
    power_law_integral_flux,
    power_law_compatibility,
)


def test_one():
    """Test one case"""
    I = power_law_integral_flux(f=1, g=2)

    assert_allclose(I, 1)


def test_powerlaw_energy_flux():
    """
    Test energy flux computation for power law against numerical solution.
    """
    e1 = Quantity(1, "TeV")
    e2 = Quantity(10, "TeV")
    e = Quantity(1, "TeV")
    g = 2.3
    I = Quantity(1e-12, "cm-2 s-1")

    val = power_law_energy_flux(I=I, g=g, e=e, e1=e1, e2=e2)

    ref = Quantity(2.1615219876151536e-12, "TeV cm-2 s-1")
    assert_quantity_allclose(val, ref)


def test_e_pivot():
    """Hard-coded example from fit example in survey/spectra.
    """
    e_pivot = power_law_pivot_energy(
        energy_ref=1, f0=5.35510540e-11, d_gamma=0.0318377, cov=6.56889442e-14
    )

    assert_allclose(e_pivot, 3.3540034240210987)


def test_compatibility():
    """
    Run a test case with hardcoded numbers:

    HESS J1912+101
    1FGL 1913.7+1007c

    We use the following symbols and units:
    e = pivot energy (MeV)
    f = flux density (cm^-2 s^-1 MeV^-1)
    g = "gamma" = spectral index
    """
    # Fermi power-law parameters
    e_fermi = 1296.2734
    f_fermi = 3.791907e-12
    f_err_fermi = 5.6907235e-13
    g_fermi = 2.3759267
    g_err_fermi = 0.08453985
    par_fermi = (e_fermi, f_fermi, f_err_fermi, g_fermi, g_err_fermi)

    # HESS power-law parameters
    e_hess = 1e6
    f_hess = 3.5 * 1e-12 * 1e-6
    f_err_hess = 0.6 * 1e-12 * 1e-6
    g_hess = 2.2
    g_err_hess = 0.2
    par_hess = (e_hess, f_hess, f_err_hess, g_hess, g_err_hess)

    compatibility = power_law_compatibility(par_fermi, par_hess)

    # Note: I just put the numbers here, didn't verify them!
    assert_allclose(compatibility["g_match"], 2.0901127509816506)
    assert_allclose(compatibility["sigma_low"], -3.380819211512078)
    assert_allclose(compatibility["sigma_high"], -0.5494362450917478)
    assert_allclose(compatibility["sigma_combined"], 3.4251742624791612)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import absolute_import, division, print_function, unicode_literals
import pytest
import numpy as np
from numpy.testing import assert_allclose
from ..sparse import SparseArray
from .._sparse import merge_sparse_arrays

test_params = [(8,), (8, 16), (8, 16, 32)]


@pytest.mark.parametrize("shape", test_params)
def test_sparse_init(shape):
    v = SparseArray(shape)
    assert v.shape == shape
    assert v.size == 0

    data = np.ones(shape)
    v = SparseArray.from_array(data)
    assert_allclose(v[...], data)


def test_sparse_getitem():
    shape = (8, 16, 32)

    data = np.random.poisson(np.ones(shape)).astype(float)
    v = SparseArray.from_array(data)
    assert_allclose(v[...], data[...])
    assert_allclose(v[:], data[:])
    assert_allclose(v[:, :], data[:, :])
    assert_allclose(v[:, :, :], data[:, :, :])
    assert_allclose(v[1, 3, 10], data[1, 3, 10])
    assert_allclose(v[:, 3, 10], data[:, 3, 10])
    assert_allclose(v[1, :, 10], data[1, :, 10])
    assert_allclose(v[1, 3, :], data[1, 3, :])
    assert_allclose(v[1, np.arange(4), :], data[1, np.arange(4), :])
    assert_allclose(
        v[np.arange(4), np.arange(4), :], data[np.arange(4), np.arange(4), :]
    )
    assert_allclose(
        v[:, np.arange(4), np.arange(4)], data[:, np.arange(4), np.arange(4)]
    )


@pytest.mark.parametrize("shape", test_params)
def test_sparse_setitem(shape):
    data = np.random.poisson(np.ones(shape)).astype(float)
    v = SparseArray(shape)
    idx = np.where(data > 0)
    v[idx] = data[idx]
    assert_allclose(v[...], data)

    v = SparseArray(shape)
    v.set(idx, data[idx])
    assert_allclose(v[...], data)

    v = SparseArray(shape)
    v.set(idx, data[idx], fill=True)
    v.set(idx, data[idx], fill=True)
    assert_allclose(v[...], 2.0 * data)

    v = SparseArray(shape)
    data0 = np.random.poisson(np.ones(shape)).astype(float)
    data1 = np.random.poisson(np.ones(shape)).astype(float)
    data = data0 + data1
    idx0 = np.where(data0 > 0)
    idx1 = np.where(data1 > 0)
    idx_in = tuple([np.concatenate((x, y)) for x, y in zip(idx0, idx1)])
    data_in = np.concatenate((data0[idx0], data1[idx1]))
    v.set(idx_in, data_in, fill=True)
    assert_allclose(v[...], data)


@pytest.mark.parametrize(
    ("dtype_idx", "dtype_val"),
    [
        (np.int64, np.float64),
        (np.int32, np.float64),
        (np.int32, np.float32),
        (np.int32, np.float64),
    ],
)
def test_merge_sparse_arrays(dtype_idx, dtype_val):
    idx0 = np.array([0, 0, 1, 4], dtype=dtype_idx)
    val0 = np.array([1.0, 2.0, 3.0, 7.0], dtype=dtype_val)
    idx1 = np.array([0, 1, 2], dtype=dtype_idx)
    val1 = np.array([1.0, 1.0, 1.0], dtype=dtype_val)
    idx, val = merge_sparse_arrays(idx0, val0, idx1, val1)
    assert_allclose(idx, np.unique(np.concatenate((idx0, idx1))))
    assert_allclose(val, np.array([2.0, 3.0, 1.0, 7.0]))
    idx, val = merge_sparse_arrays(idx0, val0, idx1, val1, True)
    assert_allclose(idx, np.unique(np.concatenate((idx0, idx1))))
    assert_allclose(val, np.array([4.0, 4.0, 1.0, 7.0]))



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Example and test datasets."""
from __future__ import absolute_import, division, print_function, unicode_literals
import warnings
from astropy.units import Quantity, UnitsWarning
from astropy.io import fits
from astropy.table import Table
from .core import gammapy_data

__all__ = [
    "load_poisson_stats_image",
    "load_tev_spectrum",
    "load_crab_flux_points",
    "load_diffuse_gamma_spectrum",
    "load_electron_spectrum",
]


def load_poisson_stats_image(extra_info=False, return_filenames=False):
    """Load Poisson statistics counts image of a Gaussian source on flat background.

    See poissson_stats_image/README.md for further info.
    TODO: add better description (extract from README?)

    Parameters
    ----------
    extra_info : bool
        If true, a dict of images is returned.
    return_filenames : bool
        If true, return filenames instead of images

    Returns
    -------
    data : numpy array or dict of arrays or filenames
        Depending on the ``extra_info`` and ``return_filenames`` options.
    """
    path = gammapy_data.dir / "tests/unbundled/poisson_stats_image"

    if extra_info:
        out = dict()
        for name in ["counts", "model", "source", "background", "exposure"]:
            filename = str(path / "{}.fits.gz".format(name))
            if return_filenames:
                out[name] = filename
            else:
                data = fits.getdata(filename)
                out[name] = data.astype("float64")
        if return_filenames:
            out["psf"] = str(path / "psf.json")
    else:
        filename = str(path / "counts.fits.gz")
        if return_filenames:
            out = filename
        else:
            out = fits.getdata(filename).astype("float64")

    if extra_info and not return_filenames:
        filename = str(path / "counts.fits.gz")
        out["header"] = fits.getheader(filename)

    return out


def load_tev_spectrum(source_name):
    """Load published TeV flux point measurements.

    TODO: give references to publications and describe the returned table.

    Parameters
    ----------
    source_name : str
        Source name

    Returns
    -------
    spectrum : `~astropy.table.Table`
        Energy spectrum as a table (one flux point per row).
    """
    if source_name == "crab":
        filename = gammapy_data.filename(
            "tests/unbundled/tev_spectra/crab_hess_spec.txt"
        )
    else:
        raise ValueError("Data not available for source: {!r}".format(source_name))

    names = ["energy", "flux", "flux_lo", "flux_hi"]
    table = Table.read(filename, format="ascii", names=names)
    table["flux_err"] = 0.5 * (table["flux_lo"] + table["flux_hi"])
    return table


def load_crab_flux_points(component="both"):
    """Load published Crab pulsar and nebula flux points.

    Besides the usual flux point columns, this table contains
    the following two columns:

    * component : {'pulsar', 'nebula'}
    * paper : Short string describing which point originates from which paper.

    TODO:

    * Add link to Crab flux point tutorial in Gammapy where these points are plotted.
    * Add html links to ADS directly in the docstring and as a table column.

    Parameters
    ----------
    component : {'pulsar', 'nebula', 'both'}
        Which emission component to include

    Returns
    -------
    flux_points : `~astropy.table.Table`
        Flux point table

    Notes
    -----
    This data compilation is from Buehler and Blandford, Rep. Prog. Phys. 77, 2014.
    It was contributed to Gammapy directly by Rolf Buehler via a pull request.

    The data for the nebula were taken from Meyer et al. Astron. Astrophys. 523 2010
    with the addition of the Fermi-LAT measurement reported in Buehler et al. ApJ 749 2012.

    The pulsar spectrum is reproduced from Kuiper et al Astron. Astrophys. 378 2001 .
    Additionally shown are infrared measurements reported in Sollerman et al. ApJ 537 2000
    and Tziamtzis et al. Astron. Astrophys. 508 2009, radio measurements referenced in
    Thompson et al. ApJ 516 1999 and gamma-ray measurements referenced in
    Aleksic et al. ApJ 742 2011, Aliu et al. Science 334 2011,
    Aleksic et al. Astron. Astrophys. 540 2012
    and Abdo et al. Astrophys. J. Suppl. Ser. 208 2013.
    """
    filename = gammapy_data.filename("tests/unbundled/tev_spectra/crab_mwl.fits.gz")

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", UnitsWarning)
        table = Table.read(filename)

    if component == "pulsar":
        mask = table["component"] == "pulsar"
        table = table[mask]
    elif component == "nebula":
        mask = table["component"] == "nebula"
        table = table[mask]
    elif component == "both":
        pass
    else:
        raise ValueError("Invalid component: {}".format(component))

    return table


def load_diffuse_gamma_spectrum(reference):
    """Load published diffuse gamma-ray spectrum.

    TODO: give references to publications and describe the returned table.

    Parameters
    ----------
    reference : {'Fermi', 'Fermi2'}
        Which publication.

    Returns
    -------
    spectrum : `~astropy.table.Table`
        Energy spectrum as a table (one flux point per row).
    """
    dir = gammapy_data.dir / "tests/unbundled/tev_spectra"

    if reference == "Fermi":
        filename = str(dir / "diffuse_isotropic_gamma_spectrum_fermi.txt")
    elif reference == "Fermi2":
        filename = str(dir / "diffuse_isotropic_gamma_spectrum_fermi2.txt")
    else:
        raise ValueError("Data not available for reference: {}".format(reference))

    return _read_diffuse_gamma_spectrum_fermi(filename)


def _read_diffuse_gamma_spectrum_fermi(filename):
    table = Table.read(
        filename, format="ascii", names=["energy", "flux", "flux_hi", "flux_lo"]
    )
    table["flux_err"] = 0.5 * (table["flux_lo"] + table["flux_hi"])

    table["energy"] = Quantity(table["energy"], "MeV").to("TeV")

    for colname in table.colnames:
        if "flux" in colname:
            energy = Quantity(table["energy"], "TeV")
            energy2_flux = Quantity(table[colname], "MeV cm^-2 s^-1 sr^-1")
            table[colname] = (energy2_flux / energy ** 2).to("m^-2 s^-1 TeV^-1 sr^-1")

    return table


def load_electron_spectrum(reference):
    """Load published electron spectrum.

    TODO: give references to publications and describe the returned table.

    Parameters
    ----------
    reference : {'HESS', 'HESS low energy', 'Fermi'}
        Which publication.

    Returns
    -------
    spectrum : `~astropy.table.Table`
        Energy spectrum as a table (one flux point per row).
    """
    dir = gammapy_data.dir / "tests/unbundled/tev_spectra"

    if reference == "HESS":
        filename = str(dir / "electron_spectrum_hess.txt")
        return _read_electron_spectrum_hess(filename)
    elif reference == "HESS low energy":
        filename = str(dir / "electron_spectrum_hess_low_energy.txt")
        return _read_electron_spectrum_hess(filename)
    elif reference == "Fermi":
        filename = str(dir / "electron_spectrum_fermi.txt")
        return _read_electron_spectrum_fermi(filename)
    else:
        raise ValueError("Data not available for reference: {}".format(reference))


def _read_electron_spectrum_hess(filename):
    table = Table.read(
        filename, format="ascii", names=["energy", "flux", "flux_lo", "flux_hi"]
    )
    table["flux_err"] = 0.5 * (table["flux_lo"] + table["flux_hi"])

    table["energy"] = Quantity(table["energy"], "GeV").to("TeV")

    # The ascii files store fluxes as (E ** 3) * dN / dE.
    # Here we change this to dN / dE.
    for colname in table.colnames:
        if "flux" in colname:
            energy = Quantity(table["energy"], "TeV")
            energy3_flux = Quantity(table[colname], "GeV^2 m^-2 s^-1 sr^-1")
            table[colname] = (energy3_flux / energy ** 3).to("m^-2 s^-1 TeV^-1 sr^-1")

    return table


def _read_electron_spectrum_fermi(filename):
    t = Table.read(filename, format="ascii")

    table = Table()
    table["energy"] = Quantity(t["E"], "GeV").to("TeV")
    table["flux"] = Quantity(t["y"], "m-2 s-1 GeV-1 sr-1").to("m-2 s-1 TeV-1 sr-1")
    val = 0.5 * (t["yerrtot_lo"] + t["yerrtot_up"])
    table["flux_err"] = Quantity(val, "m-2 s-1 GeV-1 sr-1").to("m-2 s-1 TeV-1 sr-1")

    return table



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import absolute_import, division, print_function, unicode_literals
import numpy as np

__all__ = ["fov_to_sky", "sky_to_fov"]


def fov_to_sky(lon, lat, lon_pnt, lat_pnt):
    """Make a transformation from field-of-view coordinates to sky coordinates.

    Parameters
    ----------
    lon : array_like
        Field-of-view longitude coordinate to be transformed, in degrees
    lat : array_like
        Field-of-view latitude coordinate to be transformed, in degrees
    lon_pnt : array_like
        Longitude coordinate of the pointing position, in degrees
    lat_pnt : array_like
        Latitude coordinate of the pointing position, in degrees

    Returns
    -------
    lon_t : array_like
        Sky longitude coordinate, in degrees
    lat_t : array_like
        Sky latitude coordinate, in degrees
    """
    # compute cartesian coordinates
    x = np.cos(np.deg2rad(lat)) * np.cos(np.deg2rad(lon))
    y = np.cos(np.deg2rad(lat)) * np.sin(np.deg2rad(lon))
    z = np.sin(np.deg2rad(lat))

    # make sure vector is properly normalised
    assert np.allclose(x ** 2 + y ** 2 + z ** 2, 1)

    # switch coordinates due to axis convention
    x_ = -z
    y_ = y
    z_ = x

    # transform
    lat_pnt = np.deg2rad(90 - lat_pnt)
    lon_pnt = -np.deg2rad(lon_pnt)
    x_t = (
        x_ * np.cos(lat_pnt) * np.cos(lon_pnt)
        - y_ * np.sin(lon_pnt)
        + z_ * np.sin(lat_pnt) * np.cos(lon_pnt)
    )
    y_t = (
        x_ * np.sin(lon_pnt) * np.cos(lat_pnt)
        + y_ * np.cos(lon_pnt)
        + z_ * np.sin(lon_pnt) * np.sin(lat_pnt)
    )
    z_t = -x_ * np.sin(lat_pnt) + z_ * np.cos(lat_pnt)

    # compute new lon, lat
    lon_t = -np.rad2deg(np.arctan2(y_t, x_t))
    lat_t = np.rad2deg(np.arcsin(z_t))

    # shift lon by 360 degrees if negative
    lon_t = np.where(lon_t < 0, lon_t + 360, lon_t)

    return lon_t, lat_t


def sky_to_fov(lon, lat, lon_pnt, lat_pnt):
    """Make a transformation from sky coordinates to field-of-view coordinates.

    Parameters
    ----------
    lon : array_like
        Sky longitude coordinate to be transformed, in degrees
    lat : array_like
        Sky latitude coordinate to be transformed, in degrees
    lon_pnt : array_like
        Longitude coordinate of the pointing position, in degrees
    lat_pnt : array_like
        Latitude coordinate of the pointing position, in degrees

    Returns
    -------
    lon_t : array_like
        Field-of-view longitude coordinate, in degrees
    lat_t : array_like
        Field-of-view latitude coordinate, in degrees
    """
    lon_ = -lon

    # compute cartesian coordinates
    x = np.cos(np.deg2rad(lat)) * np.cos(np.deg2rad(lon_))
    y = np.cos(np.deg2rad(lat)) * np.sin(np.deg2rad(lon_))
    z = np.sin(np.deg2rad(lat))

    # make sure vector is properly normalised
    assert np.allclose(x ** 2 + y ** 2 + z ** 2, 1)

    # transform
    lat_pnt = np.deg2rad(90 - lat_pnt)
    lon_pnt = -np.deg2rad(lon_pnt)
    x_t = (
        x * np.cos(lat_pnt) * np.cos(lon_pnt)
        + y * np.sin(lon_pnt) * np.cos(lat_pnt)
        - z * np.sin(lat_pnt)
    )
    y_t = -x * np.sin(lon_pnt) + y * np.cos(lon_pnt)
    z_t = (
        x * np.sin(lat_pnt) * np.cos(lon_pnt)
        + y * np.sin(lat_pnt) * np.sin(lon_pnt)
        + z * np.cos(lat_pnt)
    )

    # switch coordinates due to axis convention
    x_ = z_t
    y_ = y_t
    z_ = -x_t

    # compute new lon, lat
    lon_t = np.rad2deg(np.arctan2(y_, x_))
    lat_t = np.rad2deg(np.arcsin(z_))

    return lon_t, lat_t



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import absolute_import, division, print_function, unicode_literals
import pytest
from numpy.testing import assert_allclose
import astropy.units as u
from ...utils.testing import requires_data, requires_dependency
from ..hawc import SourceCatalog2HWC


@pytest.fixture(scope="session")
def hawc_2hwc():
    return SourceCatalog2HWC()


@requires_data("gammapy-data")
class TestSourceCatalog2HWC:
    def test_source_table(self, hawc_2hwc):
        assert hawc_2hwc.name == "2hwc"
        assert len(hawc_2hwc.table) == 40


@requires_data("gammapy-data")
class TestSourceCatalogObject2HWC:
    def test_data(self, hawc_2hwc):
        source = hawc_2hwc[0]
        assert source.data["source_name"] == "2HWC J0534+220"

    def test_str(self, hawc_2hwc):
        source = hawc_2hwc[0]
        assert "2HWC J0534+220" in str(source)
        assert "No second spectrum available for this source" in str(source)

        source = hawc_2hwc[1]
        assert "2HWC J0631+169" in str(source)
        assert "Spectrum 1:" in str(source)

    @requires_dependency("uncertainties")
    def test_spectral_models_one(self, hawc_2hwc):
        source = hawc_2hwc[0]
        assert source.n_spectra == 1

        spectral_models = source.spectral_models
        assert len(spectral_models) == 1

        e_min, e_max = [1, 10] * u.TeV
        flux, flux_err = spectral_models[0].integral_error(e_min, e_max)
        assert flux.unit == "cm-2 s-1"
        assert_allclose(flux.value, 1.2966462620662674e-12)
        assert flux_err.unit == "cm-2 s-1"
        assert_allclose(flux_err.value, 1.671177271712936e-14)

    @requires_dependency("uncertainties")
    def test_spectral_models_two(self, hawc_2hwc):
        # This test is just to check that sources with 2 spectra also work OK.
        source = hawc_2hwc[1]
        assert source.n_spectra == 2

        spectral_models = source.spectral_models
        assert len(spectral_models) == 2

        e_min, e_max = [1, 10] * u.TeV
        flux, flux_err = spectral_models[1].integral_error(e_min, e_max)
        assert flux.unit == "cm-2 s-1"
        assert_allclose(flux.value, 3.3381204455973463e-13)
        assert flux_err.unit == "cm-2 s-1"
        assert_allclose(flux_err.value, 4.697084075095061e-14)



# Licensed under a 3 - clause BSD style license - see LICENSE.rst
from __future__ import absolute_import, division, print_function, unicode_literals
from collections import OrderedDict
import logging
from ..utils.random import get_random_state
from ..utils.energy import EnergyBounds
from .utils import CountsPredictor
from .core import PHACountsSpectrum
from .observation import SpectrumObservation, SpectrumObservationList

__all__ = ["SpectrumSimulation"]

log = logging.getLogger(__name__)


class SpectrumSimulation(object):
    """Simulate `~gammapy.spectrum.SpectrumObservation`.

    For a usage example see :gp-notebook:`spectrum_simulation`

    Parameters
    ----------
    livetime : `~astropy.units.Quantity`
        Livetime
    source_model : `~gammapy.spectrum.models.SpectralModel`
        Source model
    aeff : `~gammapy.irf.EffectiveAreaTable`, optional
        Effective Area
    edisp : `~gammapy.irf.EnergyDispersion`, optional
        Energy Dispersion
    e_true : `~astropy.units.Quantity`, optional
        Desired energy axis of the prediced counts vector if no IRFs are given
    background_model : `~gammapy.spectrum.models.SpectralModel`, optional
        Background model
    alpha : float, optional
        Exposure ratio between source and background
    """

    def __init__(
        self,
        livetime,
        source_model,
        aeff=None,
        edisp=None,
        e_true=None,
        background_model=None,
        alpha=None,
    ):
        self.livetime = livetime
        self.source_model = source_model
        self.aeff = aeff
        self.edisp = edisp
        self.e_true = e_true
        self.background_model = background_model
        self.alpha = alpha

        self.on_vector = None
        self.off_vector = None
        self.obs = None
        self.result = SpectrumObservationList()

    @property
    def npred_source(self):
        """Predicted source `~gammapy.spectrum.CountsSpectrum`.

        Calls :func:`gammapy.spectrum.utils.CountsPredictor`.
        """
        predictor = CountsPredictor(
            livetime=self.livetime,
            aeff=self.aeff,
            edisp=self.edisp,
            e_true=self.e_true,
            model=self.source_model,
        )
        predictor.run()
        return predictor.npred

    @property
    def npred_background(self):
        """Predicted background (`~gammapy.spectrum.CountsSpectrum`).

        Calls :func:`gammapy.spectrum.utils.CountsPredictor`.
        """
        predictor = CountsPredictor(
            livetime=self.livetime,
            aeff=self.aeff,
            edisp=self.edisp,
            e_true=self.e_true,
            model=self.background_model,
        )
        predictor.run()
        return predictor.npred

    @property
    def e_reco(self):
        """Reconstructed energy binning."""
        if self.edisp is not None:
            temp = self.edisp.e_reco.bins
        else:
            if self.aeff is not None:
                temp = self.aeff.energy.bins
            else:
                temp = self.e_true
        return EnergyBounds(temp)

    def run(self, seed):
        """Simulate `~gammapy.spectrum.SpectrumObservationList`.

        The seeds will be set as observation ID.
        Previously produced results will be overwritten.

        Parameters
        ----------
        seed : array of ints
            Random number generator seeds
        """
        self.reset()
        n_obs = len(seed)
        log.info("Simulating {} observations".format(n_obs))
        for counter, current_seed in enumerate(seed):
            progress = ((counter + 1) / n_obs) * 100
            if progress % 10 == 0:
                log.info("Progress : {} %".format(progress))
            self.simulate_obs(seed=current_seed, obs_id=current_seed)
            self.result.append(self.obs)

    def reset(self):
        """Clear all results."""
        self.result = SpectrumObservationList()
        self.obs = None
        self.on_vector = None
        self.off_vector = None

    def simulate_obs(self, obs_id, seed="random-seed"):
        """Simulate one `~gammapy.spectrum.SpectrumObservation`.

        The result is stored as ``obs`` attribute

        Parameters
        ----------
        obs_id : int
            Observation identifier
        seed : {int, 'random-seed', 'global-rng', `~numpy.random.RandomState`}
            see :func:~`gammapy.utils.random.get_random_state`
        """
        random_state = get_random_state(seed)
        self.simulate_source_counts(random_state)
        if self.background_model is not None:
            self.simulate_background_counts(random_state)
        obs = SpectrumObservation(
            on_vector=self.on_vector,
            off_vector=self.off_vector,
            aeff=self.aeff,
            edisp=self.edisp,
        )
        obs.obs_id = obs_id
        self.obs = obs

    def simulate_source_counts(self, rand):
        """Simulate source `~gammapy.spectrum.PHACountsSpectrum`.

        Source counts are added to the on vector.

        Parameters
        ----------
        rand : `~numpy.random.RandomState`
            random state
        """
        on_counts = rand.poisson(self.npred_source.data.data.value)

        on_vector = PHACountsSpectrum(
            energy_lo=self.e_reco.lower_bounds,
            energy_hi=self.e_reco.upper_bounds,
            data=on_counts,
            backscal=1,
            meta=self._get_meta(),
        )
        on_vector.livetime = self.livetime
        self.on_vector = on_vector

    def simulate_background_counts(self, rand):
        """Simulate background `~gammapy.spectrum.PHACountsSpectrum`.

        Background counts are added to the on vector.
        Furthermore background counts divided by alpha are added to the off vector.

        TODO: At the moment the source IRFs are used.
        Make it possible to pass dedicated background IRFs.

        Parameters
        ----------
        rand : `~numpy.random.RandomState`
            random state
        """
        bkg_counts = rand.poisson(self.npred_background.data.data.value)
        off_counts = rand.poisson(self.npred_background.data.data.value / self.alpha)

        # Add background to on_vector
        self.on_vector.data.data += bkg_counts

        # Create off vector
        off_vector = PHACountsSpectrum(
            energy_lo=self.e_reco.lower_bounds,
            energy_hi=self.e_reco.upper_bounds,
            data=off_counts,
            backscal=1.0 / self.alpha,
            is_bkg=True,
            meta=self._get_meta(),
        )
        off_vector.livetime = self.livetime
        self.off_vector = off_vector

    def _get_meta(self):
        return OrderedDict([("CREATOR", self.__class__.__name__)])



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import absolute_import, division, print_function, unicode_literals
import pytest
import numpy as np
from numpy.testing import assert_allclose
from ...testing import requires_data
from ....spectrum import models as spectral
from ....image import models as spatial
from ....cube.models import SkyModels
from ...serialization import xml_to_sky_models, UnknownModelError


def test_from_xml():
    xml = """<?xml version="1.0" encoding="utf-8"?>
        <source_library title="source library">
            <source name="3C 273" type="PointSource">
                <spectrum type="PowerLaw">;i
                    <parameter free="1" max="1000.0" min="0.001" name="Prefactor" scale="1e-09" value="10"></parameter>
                    <parameter free="1" max="-1.0" min="-5.0" name="Index" scale="1.0" value="-2.1"></parameter>
                    <parameter free="0" max="2000.0" min="30.0" name="Scale" scale="1.0" value="100.0"></parameter>
                </spectrum>
                <spatialModel type="SkyDirFunction">
                    <parameter free="0" max="360" min="-360" name="RA" scale="1.0" value="187.25"></parameter>
                    <parameter free="0" max="90" min="-90" name="DEC" scale="1.0" value="2.17"></parameter>
                </spatialModel>
            </source>
        </source_library>
        """
    sky_models = SkyModels.from_xml(xml)
    sky_model = sky_models.skymodels[0]
    assert_allclose(sky_model.parameters["lon_0"].value, 187.25)


def test_xml_errors():
    xml = '<?xml version="1.0" standalone="no"?>\n'
    xml += '<source_library title="broken source library">\n'
    xml += '    <source name="CrabShell" type="ExtendedSource">\n'
    xml += '        <spatialModel type="ElefantShapedSource">\n'
    xml += '            <parameter name="RA" value="1" scale="1" min="1" max="1" free="1"/>\n'
    xml += "        </spatialModel>\n"
    xml += "    </source>\n"
    xml += "</source_library>"

    with pytest.raises(UnknownModelError):
        xml_to_sky_models(xml)

    # TODO: Think about a more elaborate XML validation scheme


@requires_data("gammapy-data")
def test_complex():
    filename = "$GAMMAPY_DATA/tests/models/examples.xml"
    sourcelib = SkyModels.read(filename)

    assert len(sourcelib.skymodels) == 7

    model0 = sourcelib.skymodels[0]
    assert isinstance(model0.spectral_model, spectral.PowerLaw)
    assert isinstance(model0.spatial_model, spatial.SkyPointSource)

    model1 = sourcelib.skymodels[1]
    assert isinstance(model1.spectral_model, spectral.PowerLaw)
    assert isinstance(model1.spatial_model, spatial.SkyPointSource)

    pars1 = model1.parameters
    assert pars1["index"].value == 2.1
    assert pars1["index"].unit == ""
    assert pars1["index"].max is np.nan
    assert pars1["index"].min is np.nan
    assert pars1["index"].frozen is False

    assert pars1["lon_0"].value == 0.5
    assert pars1["lon_0"].unit == "deg"
    assert pars1["lon_0"].max == 360
    assert pars1["lon_0"].min == -360
    assert pars1["lon_0"].frozen is True

    assert pars1["lat_0"].value == 1.0
    assert pars1["lat_0"].unit == "deg"
    assert pars1["lat_0"].max == 90
    assert pars1["lat_0"].min == -90
    assert pars1["lat_0"].frozen is True

    model2 = sourcelib.skymodels[2]
    assert isinstance(model2.spectral_model, spectral.ExponentialCutoffPowerLaw)
    assert isinstance(model2.spatial_model, spatial.SkyGaussian)

    pars2 = model2.parameters
    assert pars2["sigma"].unit == "deg"
    assert pars2["lambda_"].value == 0.01
    assert pars2["lambda_"].unit == "MeV-1"
    assert pars2["lambda_"].min is np.nan
    assert pars2["lambda_"].max is np.nan
    assert pars2["index"].value == 2.2
    assert pars2["index"].unit == ""
    assert pars2["index"].max is np.nan
    assert pars2["index"].min is np.nan

    model3 = sourcelib.skymodels[3]
    assert isinstance(model3.spatial_model, spatial.SkyDisk)
    pars3 = model3.parameters
    assert pars3["r_0"].unit == "deg"

    model4 = sourcelib.skymodels[4]
    assert isinstance(model4.spatial_model, spatial.SkyShell)
    pars4 = model4.parameters
    assert pars4["radius"].unit == "deg"
    assert pars4["width"].unit == "deg"

    model5 = sourcelib.skymodels[5]
    assert isinstance(model5.spatial_model, spatial.SkyDiffuseMap)

    model6 = sourcelib.skymodels[6]
    assert isinstance(model6.spatial_model, spatial.SkyDiffuseMap)


@pytest.mark.xfail(reason="Need to improve XML read")
@requires_data("gammapy-data")
@pytest.mark.parametrize(
    "filename",
    [
        "$GAMMAPY_DATA/tests/models/fermi_model.xml",
        "$GAMMAPY_DATA/tests/models/shell.xml",
    ],
)
def test_models(filename, tmpdir):
    outfile = tmpdir / "models_out.xml"
    sourcelib = SkyModels.read(filename)
    sourcelib.to_xml(outfile)

    sourcelib_roundtrip = SkyModels.from_xml(outfile)

    for model, model_roundtrip in zip(
        sourcelib.skymodels, sourcelib_roundtrip.skymodels
    ):
        assert str(model) == str(model_roundtrip)


@pytest.mark.xfail(reason="Need to improve XML read")
@requires_data("gammapy-data")
def test_sky_models_old_xml_file():
    filename = "$GAMMAPY_DATA/tests/models/shell.xml"
    sources = SkyModels.read(filename)

    assert len(sources.source_list) == 2

    source = sources.source_list[0]
    assert source.source_name == "CrabShell"
    assert_allclose(source.spectral_model.parameters["Index"].value, 2.48)

    xml = sources.to_xml()
    assert "sources" in xml


@pytest.mark.xfail(reason="Need to improve XML read")
@requires_data("gammapy-data")
def test_sky_models_new_xml_file():
    filename = "$GAMMAPY_DATA/tests/models/ctadc_skymodel_gps_sources_bright.xml"
    sources = SkyModels.read(filename)

    assert len(sources.source_list) == 47

    source = sources.source_list[0]
    assert source.source_name == "HESS J0632+057"
    assert_allclose(
        source.spectral_model.parameters["Index"].value, -2.5299999713897705
    )

    xml = sources.to_xml()
    assert "sources" in xml



"""Example to show how to plot spectrum of Fermi/LAT sources.
"""
import matplotlib.pyplot as plt
from gammapy.catalog import (
    SourceCatalog3FGL,
    SourceCatalog2FHL,
    SourceCatalog1FHL,
    SourceCatalog3FHL,
)


def plot_source_spectrum(source, label, color):
    opts = dict(energy_power=2, flux_unit="erg-1 cm-2 s-1")
    source.spectral_model.plot(
        energy_range=source.energy_range, label=label, color=color, **opts
    )
    source.spectral_model.plot_error(
        energy_range=source.energy_range, facecolor=color, **opts
    )
    source.flux_points.plot(sed_type="dnde", color=color, **opts)


def plot_source_spectra(name):
    plot_source_spectrum(
        source=SourceCatalog3FGL()[name], label="Fermi 3FGL", color="r"
    )
    plot_source_spectrum(
        source=SourceCatalog2FHL()[name], label="Fermi 2FHL", color="g"
    )
    plot_source_spectrum(
        source=SourceCatalog1FHL()[name], label="Fermi 1FHL", color="c"
    )
    plot_source_spectrum(
        source=SourceCatalog3FHL()[name], label="Fermi 3FHL", color="b"
    )

    ax = plt.gca()
    ax.set_ylim(1.0e-12, 7.0e-11)
    ax.set_xlim(1.0e-4, 2.0)
    ax.set_xlabel("Energy (TeV)")
    ax.set_ylabel("E^2 dN/dE (erg cm-2 s-1])")
    plt.legend(loc=0)


if __name__ == "__main__":
    # Select your favourite source
    # (must be named like this in ASSOC columns of all catalogs)
    name = "PKS 2155-304"
    plot_source_spectra(name)
    plt.show()



# Licensed under a 3-clause BSD style license - see LICENSE.rst
"""Simple models for cosmic ray spectra at Earth.

For measurements, the "Database of Charged Cosmic Rays (CRDB)" is a great resource:
http://lpsc.in2p3.fr/cosmic-rays-db/
"""
from __future__ import absolute_import, division, print_function, unicode_literals
import numpy as np
from astropy.units import Quantity

__all__ = ["cosmic_ray_flux"]


def _power_law(E, N, k):
    E = Quantity(E, "TeV")
    E0 = Quantity(1, "TeV")
    N = Quantity(N, "m^-2 s^-1 TeV^-1 sr^-1")
    flux = N * (E / E0) ** (-k)
    return flux


def _log_normal(E, L, E_p, w):
    E = Quantity(E, "TeV")
    E_p = Quantity(E_p, "TeV")
    L = Quantity(L, "m^-2 s^-1 sr^-1")
    term1 = L / (E * w * np.sqrt(2 * np.pi))
    term2 = np.exp(-np.log(E / E_p) ** 2 / (2 * w ** 2))
    return term1 * term2


def _electron_spectrum(E, N, k, L, E_p, w):
    flux = _power_law(E, N, k)
    flux += _log_normal(E, L, E_p, w)
    return flux


def cosmic_ray_flux(energy, particle="proton"):
    """Cosmic ray flux at Earth.

    These are the spectra assumed in this CTA study:
    Table 3 in http://adsabs.harvard.edu/abs/2013APh....43..171B

    The hadronic spectra are simple power-laws, the electron spectrum
    is the sum of  a power law and a log-normal component to model the
    "Fermi shoulder".

    Parameters
    ----------
    energy : `~astropy.units.Quantity`
        Particle energy
    particle : {'electron', 'proton', 'He', 'N', 'Si', 'Fe'}
        Particle type

    Returns
    -------
    flux : `~astropy.units.Quantity`
        Cosmic ray flux in unit ``m^-2 s^-1 TeV^-1 sr^-1``
    """
    pars = {
        "electron": {"N": 6.85e-5, "k": 3.21, "L": 3.19e-3, "E_p": 0.107, "w": 0.776},
        "proton": {"N": 0.096, "k": 2.70},
        "N": {"N": 0.0719, "k": 2.64},
        "Si": {"N": 0.0284, "k": 2.66},
        "Fe": {"N": 0.0134, "k": 2.63},
    }

    if particle == "electron":
        return _electron_spectrum(energy, **pars["electron"])
    elif particle in ["proton", "He", "N", "Si", "Fe"]:
        return _power_law(energy, **pars[particle])
    else:
        raise ValueError("Invalid argument for particle: {}".format(particle))



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import absolute_import, division, print_function, unicode_literals
import logging
import numpy as np
from scipy.ndimage import distance_transform_edt
from astropy.coordinates import Angle
from regions import PixCoord, CirclePixelRegion
from ..maps import WcsNDMap
from .background_estimate import BackgroundEstimate

__all__ = ["ReflectedRegionsFinder", "ReflectedRegionsBackgroundEstimator"]

log = logging.getLogger(__name__)


def _compute_distance_image(mask_map):
    """Distance to nearest exclusion region.

    Compute distance image, i.e. the Euclidean (=Cartesian 2D)
    distance (in pixels) to the nearest exclusion region.

    We need to call distance_transform_edt twice because it only computes
    dist for pixels outside exclusion regions, so to get the
    distances for pixels inside we call it on the inverted mask
    and then combine both distance images into one, using negative
    distances (note the minus sign) for pixels inside exclusion regions.

    If data consist only of ones, it'll be supposed to be far away
    from zero pixels, so in capacity of answer it should be return
    the matrix with the shape as like as data but packed by constant
    value Max_Value (MAX_VALUE = 1e10).

    If data consist only of zeros, it'll be supposed to be deep inside
    an exclusion region, so in capacity of answer it should be return
    the matrix with the shape as like as data but packed by constant
    value -Max_Value (MAX_VALUE = 1e10).

    Returns
    -------
    distance : `~gammapy.maps.WcsNDMap`
        Map of distance to nearest exclusion region.
    """
    max_value = 1e10

    if np.all(mask_map.data == 1):
        dist_map = mask_map.copy(data=mask_map.data * max_value)
        return dist_map

    if np.all(mask_map.data == 0):
        dist_map = mask_map.copy(data=mask_map.data - max_value)
        return dist_map

    distance_outside = distance_transform_edt(mask_map.data)

    invert_mask = np.invert(np.array(mask_map.data, dtype=np.bool))
    distance_inside = distance_transform_edt(invert_mask)

    distance = np.where(
        mask_map.data,
        distance_outside,
        -distance_inside,  # pylint:disable=invalid-unary-operand-type
    )

    return mask_map.copy(data=distance)


class ReflectedRegionsFinder(object):
    """Find reflected regions.

    This class is responsible for placing :ref:`region_reflected` for a given
    input region and pointing position. It converts to pixel coordinates
    internally. At the moment it works only for circles. If you want to make a
    background estimate for an IACT observation using the reflected regions
    method, see also `~gammapy.background.ReflectedRegionsBackgroundEstimator`

    Parameters
    ----------
    region : `~regions.CircleSkyRegion`
        Region to rotate
    center : `~astropy.coordinates.SkyCoord`
        Rotation point
    angle_increment : `~astropy.coordinates.Angle`, optional
        Rotation angle applied when a region falls in an excluded region.
    min_distance : `~astropy.coordinates.Angle`, optional
        Minimal distance between to reflected regions
    min_distance_input : `~astropy.coordinates.Angle`, optional
        Minimal distance from input region
    max_region_number : int, optional
        Maximum number of regions to use
    exclusion_mask : `~gammapy.maps.WcsNDMap`, optional
        Exclusion mask

    Examples
    --------
    >>> from astropy.coordinates import SkyCoord, Angle
    >>> from regions import CircleSkyRegion
    >>> from gammapy.background import ReflectedRegionsFinder
    >>> pointing = SkyCoord(83.2, 22.7, unit='deg', frame='icrs')
    >>> target_position = SkyCoord(80.2, 23.5, unit='deg', frame='icrs')
    >>> theta = Angle(0.4, 'deg')
    >>> on_region = CircleSkyRegion(target_position, theta)
    >>> finder = ReflectedRegionsFinder(min_distance_input='1 rad', region=on_region, center=pointing)
    >>> finder.run()
    >>> print(finder.reflected_regions[0])
    Region: CircleSkyRegion
    center: <SkyCoord (Galactic): (l, b) in deg
        ( 184.9367087, -8.37920222)>
        radius: 0.400147197682 deg
    """

    def __init__(
        self,
        region,
        center,
        angle_increment="0.1 rad",
        min_distance="0 rad",
        min_distance_input="0.1 rad",
        max_region_number=10000,
        exclusion_mask=None,
    ):
        self.region = region
        self.center = center

        self.angle_increment = Angle(angle_increment)
        if self.angle_increment <= Angle(0, "deg"):
            raise ValueError("angle_increment is too small")

        self.min_distance = Angle(min_distance)
        self.min_distance_input = Angle(min_distance_input)
        self.exclusion_mask = exclusion_mask
        self.max_region_number = max_region_number
        self.reflected_regions = None

    def run(self):
        """Run all steps.
        """
        if self.exclusion_mask is None:
            self.exclusion_mask = self.make_empty_mask(self.region, self.center)
        self.setup()
        self.find_regions()

    @staticmethod
    def make_empty_mask(region, center):
        """Create empty exclusion mask.

        The size of the mask is chosen such that all reflected region are
        contained on the image.

        Parameters
        ----------
        region : `~regions.CircleSkyRegion`
            Region to rotate
        center : `~astropy.coordinates.SkyCoord`
            Rotation point
        """
        log.debug("No exclusion mask provided, creating an emtpy one")
        binsz = 0.02
        width = 3 * region.center.separation(center)

        maskmap = WcsNDMap.create(
            skydir=center, binsz=binsz, width=width, coordsys="GAL", proj="TAN"
        )
        maskmap.data += 1.0
        return maskmap

    def setup(self):
        """Compute parameters for reflected regions algorithm."""
        wcs = self.exclusion_mask.geom.wcs
        self._pix_region = self.region.to_pixel(wcs)
        self._pix_center = PixCoord(*self.center.to_pixel(wcs))
        dx = self._pix_region.center.x - self._pix_center.x
        dy = self._pix_region.center.y - self._pix_center.y

        # Offset of region in pix coordinates
        self._offset = np.hypot(dx, dy)

        # Starting angle of region
        self._angle = Angle(np.arctan2(dx, dy), "rad")

        # Minimum angle a circle has to be moved to not overlap with previous one
        min_ang = Angle(2 * np.arcsin(self._pix_region.radius / self._offset), "rad")

        # Add required minimal distance between two off regions
        self._min_ang = min_ang + self.min_distance

        # Maximum possible angle before regions is reached again
        self._max_angle = (
            self._angle + Angle("360deg") - self._min_ang - self.min_distance_input
        )

        # Distance image
        self._distance_image = _compute_distance_image(self.exclusion_mask)

    def find_regions(self):
        """Find reflected regions."""
        curr_angle = self._angle + self._min_ang + self.min_distance_input
        reflected_regions = []
        while curr_angle < self._max_angle:
            test_pos = self._compute_xy(self._pix_center, self._offset, curr_angle)
            test_reg = CirclePixelRegion(test_pos, self._pix_region.radius)
            if not self._is_inside_exclusion(test_reg, self._distance_image):
                refl_region = test_reg.to_sky(self.exclusion_mask.geom.wcs)
                log.debug("Placing reflected region\n{}".format(refl_region))
                reflected_regions.append(refl_region)
                curr_angle = curr_angle + self._min_ang
                if self.max_region_number <= len(reflected_regions):
                    break
            else:
                curr_angle = curr_angle + self.angle_increment

        log.debug("Found {} reflected regions".format(len(reflected_regions)))
        self.reflected_regions = reflected_regions

    def plot(self, fig=None, ax=None):
        """Standard debug plot.

        See example here: :ref:'regions_reflected'.
        """
        fig, ax, cbar = self.exclusion_mask.plot(fig=fig, ax=ax, cmap="gray")
        wcs = self.exclusion_mask.geom.wcs
        on_patch = self.region.to_pixel(wcs=wcs).as_artist(color="red", alpha=0.6)
        ax.add_patch(on_patch)

        for off in self.reflected_regions:
            tmp = off.to_pixel(wcs=wcs)
            off_patch = tmp.as_artist(color="blue", alpha=0.6)
            ax.add_patch(off_patch)

            test_pointing = self.center
            ax.scatter(
                test_pointing.galactic.l.degree,
                test_pointing.galactic.b.degree,
                transform=ax.get_transform("galactic"),
                marker="+",
                s=300,
                linewidths=3,
                color="green",
            )

        return fig, ax

    @staticmethod
    def _is_inside_exclusion(pixreg, distance_image):
        """Test if a `~regions.PixRegion` overlaps with an exclusion mask.

        If the regions is outside the exclusion mask, return 'False'
        """
        x, y = pixreg.center.x, pixreg.center.y
        try:
            val = distance_image.data[np.round(y).astype(int), np.round(x).astype(int)]
        except IndexError:
            return False
        else:
            return val < pixreg.radius

    @staticmethod
    def _compute_xy(pix_center, offset, angle):
        """Compute x, y position for a given position angle and offset.

        # TODO: replace by calculation using `astropy.coordinates`
        """
        dx = offset * np.sin(angle)
        dy = offset * np.cos(angle)
        x = pix_center.x + dx
        y = pix_center.y + dy
        return PixCoord(x=x, y=y)


class ReflectedRegionsBackgroundEstimator(object):
    """Reflected Regions background estimator.

    This class is responsible for creating a
    `~gammapy.background.BackgroundEstimate` by placing reflected regions given
    a target region and an observation.

    For a usage example see :gp-notebook:`spectrum_analysis`

    Parameters
    ----------
    on_region : `~regions.CircleSkyRegion`
        Target region
    observations : `~gammapy.data.Observations`
        Observations to process
    kwargs : dict
        Forwarded to `gammapy.background.ReflectedRegionsFinder`
    """

    def __init__(self, on_region, observations, **kwargs):
        self.on_region = on_region
        self.observations = observations
        self.finder = ReflectedRegionsFinder(region=on_region, center=None, **kwargs)

        self.result = None

    def __str__(self):
        s = self.__class__.__name__
        s += "\n{}".format(self.on_region)
        s += "\n{}".format(self.observations)
        s += "\n{}".format(self.finder)
        return s

    def run(self):
        """Run all steps."""
        log.debug("Computing reflected regions")
        result = []
        for obs in self.observations:
            temp = self.process(obs=obs)
            result.append(temp)

        self.result = result

    def process(self, obs):
        """Estimate background for one observation."""
        log.debug("Processing observation {}".format(obs))
        self.finder.center = obs.pointing_radec
        self.finder.run()
        off_region = self.finder.reflected_regions
        off_events = obs.events.select_circular_region(off_region)
        on_events = obs.events.select_circular_region(self.on_region)
        a_on = 1
        a_off = len(off_region)
        return BackgroundEstimate(
            on_region=self.on_region,
            on_events=on_events,
            off_region=off_region,
            off_events=off_events,
            a_on=a_on,
            a_off=a_off,
            method="Reflected Regions",
        )

    def plot(self, fig=None, ax=None, cmap=None, idx=None, add_legend=False):
        """Standard debug plot.

        Parameters
        ----------
        cmap : `~matplotlib.colors.ListedColormap`, optional
            Color map to use
        idx : int, optional
            Observations to include in the plot, default: all
        add_legend : boolean, optional
            Enable/disable legend in the plot, default: False
        """
        import matplotlib.pyplot as plt

        fig, ax, cbar = self.finder.exclusion_mask.plot(fig=fig, ax=ax)

        wcs = self.finder.exclusion_mask.geom.wcs
        on_patch = self.on_region.to_pixel(wcs=wcs).as_artist(color="red")
        ax.add_patch(on_patch)

        result = self.result
        obs_list = list(self.observations)
        if idx is not None:
            obs_list = np.asarray(self.observations)[idx]
            obs_list = np.atleast_1d(obs_list)
            result = np.asarray(self.result)[idx]
            result = np.atleast_1d(result)

        cmap = cmap or plt.get_cmap("viridis")
        colors = cmap(np.linspace(0, 1, len(self.observations)))

        handles = []
        for idx_ in np.arange(len(obs_list)):
            obs = obs_list[idx_]

            off_regions = result[idx_].off_region
            for off in off_regions:
                off_patch = off.to_pixel(wcs=wcs).as_artist(
                    alpha=0.8, color=colors[idx_], label="Obs {}".format(obs.obs_id)
                )
                handle = ax.add_patch(off_patch)
            if off_regions:
                handles.append(handle)

            test_pointing = obs.pointing_radec.galactic
            ax.scatter(
                test_pointing.l.degree,
                test_pointing.b.degree,
                transform=ax.get_transform("galactic"),
                marker="+",
                color=colors[idx_],
                s=300,
                linewidths=3,
            )

        if add_legend:
            ax.legend(handles=handles)

        return fig, ax, cbar



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import absolute_import, division, print_function, unicode_literals
from astropy.coordinates import Angle
from ...utils.testing import assert_quantity_allclose
from ...utils.wcs import linear_wcs_to_arrays, linear_arrays_to_wcs


def test_wcs_object():
    bins_x = Angle([1.0, 2.0, 3.0, 4.0], "deg")
    bins_y = Angle([-1.5, 0.0, 1.5], "deg")
    wcs = linear_arrays_to_wcs("X", "Y", bins_x, bins_y)
    nbins_x = len(bins_x) - 1
    nbins_y = len(bins_y) - 1
    reco_bins_x, reco_bins_y = linear_wcs_to_arrays(wcs, nbins_x, nbins_y)

    # test: reconstructed bins should match original bins
    assert_quantity_allclose(reco_bins_x, bins_x)
    assert_quantity_allclose(reco_bins_y, bins_y)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import absolute_import, division, print_function, unicode_literals
import sys
from datetime import datetime, timedelta
import pytest
import numpy as np
from numpy.testing import assert_allclose
from astropy.coordinates import SkyCoord, Angle
from astropy.time import Time
import astropy.units as u
from astropy.table import Table, Column
from regions import CircleSkyRegion
from ...utils.testing import requires_data, requires_dependency, mpl_plot_check
from ...utils.testing import assert_quantity_allclose
from ...utils.energy import EnergyBounds
from ...data import DataStore
from ...spectrum import SpectrumExtraction
from ...spectrum.models import PowerLaw
from ...background import ReflectedRegionsBackgroundEstimator
from ..lightcurve import LightCurve, LightCurveEstimator


# time time_min time_max flux flux_err flux_ul
# 48705.1757 48705.134 48705.2174 0.57 0.29 nan
# 48732.89195 48732.8503 48732.9336 0.39 0.29 nan
# 48734.0997 48734.058 48734.1414 0.48 0.29 nan
# 48738.98535 48738.9437 48739.027 nan nan 0.97
# 48741.0259 48740.9842 48741.0676 0.34 0.29 nan


@pytest.fixture(scope="session")
def lc():
    # table = Table()
    # time_ref = Time('2010-01-01')
    # meta = time_ref_to_dict(time_ref)

    meta = dict(TIMESYS="utc")

    table = Table(
        meta=meta,
        data=[
            Column(Time(["2010-01-02", "2010-01-05"]).mjd, "time"),
            Column(Time(["2010-01-01", "2010-01-03"]).mjd, "time_min"),
            Column(Time(["2010-01-03", "2010-01-10"]).mjd, "time_max"),
            Column([1e-11, 3e-11], "flux", unit="cm-2 s-1"),
            Column([0.1e-11, 0.3e-11], "flux_err", unit="cm-2 s-1"),
            Column([np.nan, 3.6e-11], "flux_ul", unit="cm-2 s-1"),
            Column([False, True], "is_ul"),
        ],
    )

    return LightCurve(table=table)


def test_lightcurve_repr(lc):
    assert repr(lc) == "LightCurve(len=2)"


def test_lightcurve_properties_time(lc):
    assert lc.time_scale == "utc"
    assert lc.time_format == "mjd"

    # Time-related attributes
    time = lc.time
    assert time.scale == "utc"
    assert time.format == "mjd"
    assert_allclose(time.mjd, [55198, 55201])

    assert_allclose(lc.time_min.mjd, [55197, 55199])
    assert_allclose(lc.time_max.mjd, [55199, 55206])

    # Note: I'm not sure why the time delta has this scale and format
    time_delta = lc.time_delta
    assert time_delta.scale == "tai"
    assert time_delta.format == "jd"
    assert_allclose(time_delta.jd, [2, 7])


def test_lightcurve_properties_flux(lc):
    flux = lc.table["flux"].quantity
    assert flux.unit == "cm-2 s-1"
    assert_allclose(flux.value, [1e-11, 3e-11])


# TODO: extend these tests to cover other time scales.
# In those cases, CSV should not round-trip because there
# is no header info in CSV to store the time scale!


@pytest.mark.skipif(
    sys.version_info >= (3, 7),
    reason="https://github.com/astropy/astropy/issues/7744#issuecomment-419813519",
)
@pytest.mark.parametrize("format", ["fits", "ascii.ecsv", "ascii.csv"])
def test_lightcurve_read_write(tmpdir, lc, format):
    filename = str(tmpdir / "spam")

    lc.write(filename, format=format)
    lc = LightCurve.read(filename, format=format)

    # Check if time-related info round-trips
    time = lc.time
    assert time.scale == "utc"
    assert time.format == "mjd"
    assert_allclose(time.mjd, [55198, 55201])


def test_lightcurve_fvar(lc):
    fvar, fvar_err = lc.compute_fvar()
    assert_allclose(fvar, 0.6982120021884471)
    # Note: the following tolerance is very low in the next assert,
    # because results differ by ~ 1e-3 between different machines
    assert_allclose(fvar_err, 0.07905694150420949, rtol=1e-2)


def test_lightcurve_chisq(lc):
    chi2, pval = lc.compute_chisq()
    assert_quantity_allclose(chi2, 1.0000000000000001e-11)
    assert_quantity_allclose(pval, 0.999997476867478)


@requires_dependency("matplotlib")
def test_lightcurve_plot(lc):
    with mpl_plot_check():
        lc.plot()


@pytest.mark.parametrize("flux_unit", ["cm-2 s-1"])
def test_lightcurve_plot_flux(lc, flux_unit):
    f, ferr = lc._get_fluxes_and_errors(flux_unit)
    assert_allclose(f, [1e-11, 3e-11])
    assert_allclose(ferr, ([0.1e-11, 0.3e-11], [0.1e-11, 0.3e-11]))


@pytest.mark.parametrize("flux_unit", ["cm-2 s-1"])
def test_lightcurve_plot_flux_ul(lc, flux_unit):
    is_ul, ful = lc._get_flux_uls(flux_unit)
    assert_allclose(is_ul, [False, True])
    assert_allclose(ful, [np.nan, 3.6e-11])


@requires_dependency("matplotlib")
@pytest.mark.parametrize(
    "time_format, output",
    [
        (
            "mjd",
            (np.array([55198.0, 55201]), (np.array([1.0, 2.0]), np.array([1.0, 5.0]))),
        ),
        (
            "iso",
            (
                np.array([datetime(2010, 1, 2), datetime(2010, 1, 5)]),
                (
                    np.array([timedelta(1), timedelta(2)]),
                    np.array([timedelta(1), timedelta(5)]),
                ),
            ),
        ),
        ("unsupported", ValueError),
    ],
)
def test_lightcurve_plot_time(lc, time_format, output):
    try:
        t, terr = lc._get_times_and_errors(time_format)
    except output:
        return
    assert np.array_equal(t, output[0])
    assert np.array_equal(terr, output[1])


# TODO: Reuse fixtures from spectrum tests
@pytest.fixture(scope="session")
def spec_extraction():
    data_store = DataStore.from_dir("$GAMMAPY_DATA/hess-dl3-dr1/")
    obs_ids = [23523, 23526]
    observations = data_store.get_observations(obs_ids)

    target_position = SkyCoord(ra=83.63308, dec=22.01450, unit="deg")
    on_region_radius = Angle("0.11 deg")
    on_region = CircleSkyRegion(center=target_position, radius=on_region_radius)

    bkg_estimator = ReflectedRegionsBackgroundEstimator(
        on_region=on_region, observations=observations
    )
    bkg_estimator.run()

    e_reco = EnergyBounds.equal_log_spacing(0.2, 100, 50, unit="TeV")  # fine binning
    e_true = EnergyBounds.equal_log_spacing(0.05, 100, 200, unit="TeV")
    extraction = SpectrumExtraction(
        observations=observations,
        bkg_estimate=bkg_estimator.result,
        containment_correction=False,
        e_reco=e_reco,
        e_true=e_true,
    )
    extraction.run()
    extraction.compute_energy_threshold(method_lo="area_max", area_percent_lo=10.0)
    return extraction


@requires_data("gammapy-data")
def test_lightcurve_estimator(spec_extraction):
    lc_estimator = LightCurveEstimator(spec_extraction)

    # param
    intervals = []
    for obs in spec_extraction.observations:
        intervals.append([obs.events.time[0], obs.events.time[-1]])

    model = PowerLaw(
        index=2.3 * u.Unit(""),
        amplitude=3.4e-11 * u.Unit("1 / (cm2 s TeV)"),
        reference=1 * u.TeV,
    )

    lc = lc_estimator.light_curve(
        time_intervals=intervals, spectral_model=model, energy_range=[0.5, 100] * u.TeV
    )
    table = lc.table

    assert isinstance(lc.table["time_min"][0], type(intervals[0][0].value))

    assert_quantity_allclose(len(table), 2)

    assert_allclose(table["flux"][0], 4.309371e-11, rtol=1e-3)
    assert_allclose(table["flux"][-1], 3.543012e-11, rtol=1e-3)

    assert_allclose(table["flux_err"][0], 4.135581e-12, rtol=1e-3)
    assert_allclose(table["flux_err"][-1], 3.654781e-12, rtol=1e-3)

    # TODO: change dataset and also add LC point with weak signal
    # or even negative excess that is an UL
    assert_allclose(table["flux_ul"][0], 5.550045e-11, rtol=1e-3)
    assert not table["is_ul"][0]

    # same but with threshold equal to 2 TeV
    lc = lc_estimator.light_curve(
        time_intervals=intervals, spectral_model=model, energy_range=[2, 100] * u.TeV
    )
    table = lc.table

    assert_allclose(table["flux"][0], 5.051995e-12, rtol=1e-3)

    # TODO: add test exercising e_reco selection
    # TODO: add asserts on all measured quantities


@requires_data("gammapy-data")
def test_lightcurve_interval_maker(spec_extraction):
    table = LightCurveEstimator.make_time_intervals_fixes(500, spec_extraction)
    intervals = list(zip(table["t_start"], table["t_stop"]))

    assert len(intervals) == 9
    t = intervals[0]
    assert_allclose(t[1].value - t[0].value, 500 / (24 * 3600), rtol=1e-5)


@requires_data("gammapy-data")
def test_lightcurve_adaptative_interval_maker(spec_extraction):
    lc_estimator = LightCurveEstimator(spec_extraction)
    separator = [
        Time((53343.94050200008 + 53343.952979345195) / 2, scale="tt", format="mjd")
    ]
    table = lc_estimator.make_time_intervals_min_significance(
        significance=3,
        significance_method="lima",
        energy_range=[0.2, 100] * u.TeV,
        spectrum_extraction=spec_extraction,
        separators=separator,
    )
    assert_allclose(table["significance"] >= 3, True)
    assert_allclose(table["t_start"][5].value, 53343.927374, rtol=1e-10)
    assert_allclose(table["alpha"][5], 0.0833333, rtol=1e-5)
    assert len(table) == 57
    assert_allclose(table["t_start"][0].value, 53343.922392, rtol=1e-10)
    assert_allclose(table["t_stop"][-1].value, 53343.973528, rtol=1e-10)
    val = (table["t_start"] < separator[0]) & (table["t_stop"] > separator[0])
    assert_allclose(val, False)



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import absolute_import, division, print_function, unicode_literals
import numpy as np
import logging
from astropy.io import fits
from astropy.table import Table
from astropy.units import Quantity, Unit
from astropy.coordinates import Angle
from astropy.convolution import Gaussian2DKernel
from astropy.stats import gaussian_fwhm_to_sigma
from ..extern.validator import validate_physical_type
from ..utils.array import array_stats_str
from ..utils.energy import Energy, EnergyBounds
from ..utils.scripts import make_path
from ..utils.gauss import MultiGauss2D
from .psf_3d import PSF3D
from . import EnergyDependentTablePSF

__all__ = ["EnergyDependentMultiGaussPSF"]

log = logging.getLogger(__name__)


class EnergyDependentMultiGaussPSF(object):
    """
    Triple Gauss analytical PSF depending on energy and theta.

    To evaluate the PSF call the ``to_energy_dependent_table_psf`` or ``psf_at_energy_and_theta`` methods.

    Parameters
    ----------
    energy_lo : `~astropy.units.Quantity`
        Lower energy boundary of the energy bin.
    energy_hi : `~astropy.units.Quantity`
        Upper energy boundary of the energy bin.
    theta : `~astropy.units.Quantity`
        Center values of the theta bins.
    sigmas : list of 'numpy.ndarray'
        Triple Gauss sigma parameters, where every entry is
        a two dimensional 'numpy.ndarray' containing the sigma
        value for every given energy and theta.
    norms : list of 'numpy.ndarray'
        Triple Gauss norm parameters, where every entry is
        a two dimensional 'numpy.ndarray' containing the norm
        value for every given energy and theta. Norm corresponds
        to the value of the Gaussian at theta = 0.
    energy_thresh_lo : `~astropy.units.Quantity`
        Lower save energy threshold of the psf.
    energy_thresh_hi : `~astropy.units.Quantity`
        Upper save energy threshold of the psf.

    Examples
    --------
    Plot R68 of the PSF vs. theta and energy:

    .. plot::
        :include-source:

        import matplotlib.pyplot as plt
        from gammapy.irf import EnergyDependentMultiGaussPSF
        filename = '$GAMMAPY_DATA/tests/unbundled/irfs/psf.fits'
        psf = EnergyDependentMultiGaussPSF.read(filename, hdu='POINT SPREAD FUNCTION')
        psf.plot_containment(0.68, show_safe_energy=False)
        plt.show()
    """

    def __init__(
        self,
        energy_lo,
        energy_hi,
        theta,
        sigmas,
        norms,
        energy_thresh_lo=Quantity(0.1, "TeV"),
        energy_thresh_hi=Quantity(100, "TeV"),
    ):

        # Validate input
        validate_physical_type("energy_lo", energy_lo, "energy")
        validate_physical_type("energy_hi", energy_hi, "energy")
        validate_physical_type("theta", theta, "angle")
        validate_physical_type("energy_thresh_lo", energy_thresh_lo, "energy")
        validate_physical_type("energy_thresh_hi", energy_thresh_hi, "energy")

        # Set attributes
        self.energy_lo = energy_lo.to("TeV")
        self.energy_hi = energy_hi.to("TeV")
        ebounds = EnergyBounds.from_lower_and_upper_bounds(
            self.energy_lo, self.energy_hi
        )
        self.energy = ebounds.log_centers
        self.theta = theta.to("deg")
        sigmas[0][sigmas[0] == 0] = 1
        sigmas[1][sigmas[1] == 0] = 1
        sigmas[2][sigmas[2] == 0] = 1
        self.sigmas = sigmas

        self.norms = norms
        self.energy_thresh_lo = energy_thresh_lo.to("TeV")
        self.energy_thresh_hi = energy_thresh_hi.to("TeV")

    @classmethod
    def read(cls, filename, hdu="PSF_2D_GAUSS"):
        """Create `EnergyDependentMultiGaussPSF` from FITS file.

        Parameters
        ----------
        filename : str
            File name
        """
        filename = make_path(filename)
        with fits.open(str(filename), memmap=False) as hdulist:
            psf = cls.from_fits(hdulist[hdu])

        return psf

    @classmethod
    def from_fits(cls, hdu):
        """Create `EnergyDependentMultiGaussPSF` from HDU list.

        Parameters
        ----------
        hdu : `~astropy.io.fits.BintableHDU`
            HDU
        """
        energy_lo = Quantity(hdu.data["ENERG_LO"][0], "TeV")
        energy_hi = Quantity(hdu.data["ENERG_HI"][0], "TeV")
        theta = Angle(hdu.data["THETA_LO"][0], "deg")

        # Get sigmas
        shape = (len(theta), len(energy_hi))
        sigmas = []
        for key in ["SIGMA_1", "SIGMA_2", "SIGMA_3"]:
            sigma = hdu.data[key].reshape(shape).copy()
            sigmas.append(sigma)

        # Get amplitudes
        norms = []
        for key in ["SCALE", "AMPL_2", "AMPL_3"]:
            norm = hdu.data[key].reshape(shape).copy()
            norms.append(norm)

        opts = {}
        try:
            opts["energy_thresh_lo"] = Quantity(hdu.header["LO_THRES"], "TeV")
            opts["energy_thresh_hi"] = Quantity(hdu.header["HI_THRES"], "TeV")
        except KeyError:
            pass

        return cls(energy_lo, energy_hi, theta, sigmas, norms, **opts)

    def to_fits(self):
        """
        Convert psf table data to FITS hdu list.

        Returns
        -------
        hdu_list : `~astropy.io.fits.HDUList`
            PSF in HDU list format.
        """
        # Set up data
        names = [
            "ENERG_LO",
            "ENERG_HI",
            "THETA_LO",
            "THETA_HI",
            "SCALE",
            "SIGMA_1",
            "AMPL_2",
            "SIGMA_2",
            "AMPL_3",
            "SIGMA_3",
        ]
        units = ["TeV", "TeV", "deg", "deg", "", "deg", "", "deg", "", "deg"]

        data = [
            self.energy_lo,
            self.energy_hi,
            self.theta,
            self.theta,
            self.norms[0],
            self.sigmas[0],
            self.norms[1],
            self.sigmas[1],
            self.norms[2],
            self.sigmas[2],
        ]

        table = Table()
        for name_, data_, unit_ in zip(names, data, units):
            table[name_] = [data_]
            table[name_].unit = unit_

        # Create hdu and hdu list
        hdu = fits.BinTableHDU(table)
        hdu.header["LO_THRES"] = self.energy_thresh_lo.value
        hdu.header["HI_THRES"] = self.energy_thresh_hi.value

        return fits.HDUList([fits.PrimaryHDU(), hdu])

    def write(self, filename, *args, **kwargs):
        """Write PSF to FITS file.

        Calls `~astropy.io.fits.HDUList.writeto`, forwarding all arguments.
        """
        self.to_fits().writeto(filename, *args, **kwargs)

    def psf_at_energy_and_theta(self, energy, theta):
        """
        Get `~gammapy.image.models.MultiGauss2D` model for given energy and theta.

        No interpolation is used.

        Parameters
        ----------
        energy : `~astropy.units.Quantity`
            Energy at which a PSF is requested.
        theta : `~astropy.coordinates.Angle`
            Offset angle at which a PSF is requested.

        Returns
        -------
        psf : `~gammapy.morphology.MultiGauss2D`
            Multigauss PSF object.
        """
        energy = Energy(energy)
        theta = Angle(theta)

        # Find nearest energy value
        i = np.argmin(np.abs(self.energy - energy))
        j = np.argmin(np.abs(self.theta - theta))

        # TODO: Use some kind of interpolation to get PSF
        # parameters for every energy and theta

        # Select correct gauss parameters for given energy and theta
        sigmas = [_[j][i] for _ in self.sigmas]
        norms = [_[j][i] for _ in self.norms]

        pars = {}
        pars["scale"], pars["A_2"], pars["A_3"] = norms
        pars["sigma_1"], pars["sigma_2"], pars["sigma_3"] = sigmas
        psf = HESSMultiGaussPSF(pars)
        return psf.to_MultiGauss2D(normalize=True)

    def containment_radius(self, energy, theta, fraction=0.68):
        """Compute containment for all energy and theta values"""
        # This is a false positive from pylint
        # See https://github.com/PyCQA/pylint/issues/2435
        energy = Energy(energy).flatten()  # pylint:disable=assignment-from-no-return
        theta = Angle(theta).flatten()
        radius = np.empty((theta.size, energy.size))

        for idx_energy in range(len(energy)):
            for idx_theta in range(len(theta)):
                try:
                    psf = self.psf_at_energy_and_theta(
                        energy[idx_energy], theta[idx_theta]
                    )
                    radius[idx_theta, idx_energy] = psf.containment_radius(fraction)
                except ValueError:
                    log.debug(
                        "Computing containment failed for E = {:.2f}"
                        " and Theta={:.2f}".format(energy[idx_energy], theta[idx_theta])
                    )
                    log.debug("Sigmas: {} Norms: {}".format(psf.sigmas, psf.norms))
                    radius[idx_theta, idx_energy] = np.nan

        return Angle(radius, "deg")

    def plot_containment(
        self, fraction=0.68, ax=None, show_safe_energy=False, add_cbar=True, **kwargs
    ):
        """
        Plot containment image with energy and theta axes.

        Parameters
        ----------
        fraction : float
            Containment fraction between 0 and 1.
        add_cbar : bool
            Add a colorbar
        """
        import matplotlib.pyplot as plt

        ax = plt.gca() if ax is None else ax

        energy = self.energy_hi
        offset = self.theta

        # Set up and compute data
        containment = self.containment_radius(energy, offset, fraction)

        # plotting defaults
        kwargs.setdefault("cmap", "GnBu")
        kwargs.setdefault("vmin", np.nanmin(containment.value))
        kwargs.setdefault("vmax", np.nanmax(containment.value))

        # Plotting
        x = energy.value
        y = offset.value
        caxes = ax.pcolormesh(x, y, containment.value, **kwargs)

        # Axes labels and ticks, colobar
        ax.semilogx()
        ax.set_ylabel("Offset ({unit})".format(unit=offset.unit))
        ax.set_xlabel("Energy ({unit})".format(unit=energy.unit))
        ax.set_xlim(x.min(), x.max())
        ax.set_ylim(y.min(), y.max())

        if show_safe_energy:
            self._plot_safe_energy_range(ax)

        if add_cbar:
            label = "Containment radius R{0:.0f} ({1})".format(
                100 * fraction, containment.unit
            )
            ax.figure.colorbar(caxes, ax=ax, label=label)

        return ax

    def _plot_safe_energy_range(self, ax):
        """add safe energy range lines to the plot"""
        esafe = self.energy_thresh_lo
        omin = self.offset.value.min()
        omax = self.offset.value.max()
        ax.hlines(y=esafe.value, xmin=omin, xmax=omax)
        label = "Safe energy threshold: {0:3.2f}".format(esafe)
        ax.text(x=0.1, y=0.9 * esafe.value, s=label, va="top")

    def plot_containment_vs_energy(
        self, fractions=[0.68, 0.95], thetas=Angle([0, 1], "deg"), ax=None, **kwargs
    ):
        """Plot containment fraction as a function of energy.
        """
        import matplotlib.pyplot as plt

        ax = plt.gca() if ax is None else ax

        energy = Energy.equal_log_spacing(self.energy_lo[0], self.energy_hi[-1], 100)

        for theta in thetas:
            for fraction in fractions:
                radius = self.containment_radius(energy, theta, fraction).squeeze()
                label = "{} deg, {:.1f}%".format(theta, 100 * fraction)
                ax.plot(energy.value, radius.value, label=label)

        ax.semilogx()
        ax.legend(loc="best")
        ax.set_xlabel("Energy (TeV)")
        ax.set_ylabel("Containment radius (deg)")

    def peek(self, figsize=(15, 5)):
        """Quick-look summary plots."""
        import matplotlib.pyplot as plt

        fig, axes = plt.subplots(nrows=1, ncols=3, figsize=figsize)

        self.plot_containment(fraction=0.68, ax=axes[0])
        self.plot_containment(fraction=0.95, ax=axes[1])
        self.plot_containment_vs_energy(ax=axes[2])

        # TODO: implement this plot
        # psf = self.psf_at_energy_and_theta(energy='1 TeV', theta='1 deg')
        # psf.plot_components(ax=axes[2])

        plt.tight_layout()

    def info(
        self,
        fractions=[0.68, 0.95],
        energies=Quantity([1.0, 10.0], "TeV"),
        thetas=Quantity([0.0], "deg"),
    ):
        """
        Print PSF summary info.

        The containment radius for given fraction, energies and thetas is
        computed and printed on the command line.

        Parameters
        ----------
        fractions : list
            Containment fraction to compute containment radius for.
        energies : `~astropy.units.Quantity`
            Energies to compute containment radius for.
        thetas : `~astropy.units.Quantity`
            Thetas to compute containment radius for.

        Returns
        -------
        ss : string
            Formatted string containing the summary info.
        """
        ss = "\nSummary PSF info\n"
        ss += "----------------\n"
        # Summarise data members
        ss += array_stats_str(self.theta.to("deg"), "Theta")
        ss += array_stats_str(self.energy_hi, "Energy hi")
        ss += array_stats_str(self.energy_lo, "Energy lo")
        ss += "Safe energy threshold lo: {0:6.3f}\n".format(self.energy_thresh_lo)
        ss += "Safe energy threshold hi: {0:6.3f}\n".format(self.energy_thresh_hi)

        for fraction in fractions:
            containment = self.containment_radius(energies, thetas, fraction)
            for i, energy in enumerate(energies):
                for j, theta in enumerate(thetas):
                    radius = containment[j, i]
                    ss += (
                        "{0:2.0f}% containment radius at theta = {1} and "
                        "E = {2:4.1f}: {3:5.8f}\n"
                        "".format(100 * fraction, theta, energy, radius)
                    )
        return ss

    def to_energy_dependent_table_psf(self, theta=None, rad=None, exposure=None):
        """
        Convert triple Gaussian PSF ot table PSF.

        Parameters
        ----------
        theta : `~astropy.coordinates.Angle`
            Offset in the field of view. Default theta = 0 deg
        rad : `~astropy.coordinates.Angle`
            Offset from PSF center used for evaluating the PSF on a grid.
            Default offset = [0, 0.005, ..., 1.495, 1.5] deg.
        exposure : `~astropy.units.Quantity`
            Energy dependent exposure. Should be in units equivalent to 'cm^2 s'.
            Default exposure = 1.

        Returns
        -------
        tabe_psf : `~gammapy.irf.EnergyDependentTablePSF`
            Instance of `EnergyDependentTablePSF`.
        """
        # Convert energies to log center
        energies = self.energy
        # Defaults and input handling
        if theta is None:
            theta = Angle(0, "deg")
        else:
            theta = Angle(theta)

        if rad is None:
            rad = Angle(np.arange(0, 1.5, 0.005), "deg")
        else:
            rad = Angle(rad).to("deg")

        psf_value = Quantity(np.zeros((energies.size, rad.size)), "deg^-2")

        for idx, energy in enumerate(energies):
            psf_gauss = self.psf_at_energy_and_theta(energy, theta)
            psf_value[idx] = Quantity(psf_gauss(rad), "deg^-2")

        return EnergyDependentTablePSF(
            energy=energies, rad=rad, exposure=exposure, psf_value=psf_value
        )

    def to_psf3d(self, rad):
        """ Creates a PSF3D from an analytical PSF.

        Parameters
        ----------
        rad : `~astropy.units.Quantity` or `~astropy.coordinates.Angle`
            the array of position errors (rad) on which the PSF3D will be defined

        Returns
        -------
        psf3d : `~gammapy.irf.PSF3D`
            the PSF3D. It will be defined on the same energy and offset values than the input psf.
        """
        offsets = self.theta
        energy = self.energy
        energy_lo = self.energy_lo
        energy_hi = self.energy_hi
        rad_lo = rad[:-1]
        rad_hi = rad[1:]

        psf_values = np.zeros(
            (rad_lo.shape[0], offsets.shape[0], energy_lo.shape[0])
        ) * Unit("sr-1")

        for i, offset in enumerate(offsets):
            psftable = self.to_energy_dependent_table_psf(offset)
            psf_values[:, i, :] = psftable.evaluate(energy, 0.5 * (rad_lo + rad_hi)).T

        return PSF3D(
            energy_lo,
            energy_hi,
            offsets,
            rad_lo,
            rad_hi,
            psf_values,
            self.energy_thresh_lo,
            self.energy_thresh_hi,
        )


class HESSMultiGaussPSF(object):
    """Multi-Gauss PSF as represented in the HESS software.

    The 2D Gaussian is represented as a 1D exponential
    probability density function per offset angle squared:
    dp / dtheta**2 = [0]*(exp(-x/(2*[1]*[1]))+[2]*exp(-x/(2*[3]*[3]))

    @param source: either a dict of a filename

    The following two parameters control numerical
    precision / speed. Usually the defaults are fine.
    @param theta_max: Maximum offset in numerical computations
    @param npoints: Number of points in numerical computations
    @param eps: Allowed tolerance on normalization of total P to 1
    """

    def __init__(self, source):
        if isinstance(source, dict):
            # Assume source is a dict with correct format
            self.pars = source
        else:
            # Assume source is a filename with correct format
            self.pars = self._read_ascii(source)
        # Scale will be computed from normalization anyways,
        # so any default is fine here
        self.pars["scale"] = self.pars.get("scale", 1)
        # This avoids handling the first PSF as a special case
        self.pars["A_1"] = self.pars.get("A_1", 1)

    def _read_ascii(self, filename):
        """Parse file with parameters."""
        fh = open(filename)  # .readlines()
        pars = {}
        for line in fh:
            try:
                key, value = line.strip().split()[:2]
                if key.startswith("#"):
                    continue
                else:
                    pars[key] = float(value)
            except ValueError:
                pass
        fh.close()
        return pars

    def n_gauss(self):
        """Count number of Gaussians."""
        return len([_ for _ in self.pars.keys() if "sigma" in _])

    def dpdtheta2(self, theta2):
        """dp / dtheta2 at position theta2 = theta ^ 2."""
        theta2 = np.asarray(theta2, "f")
        total = np.zeros_like(theta2)
        for ii in range(1, self.n_gauss() + 1):
            A = self.pars["A_{}".format(ii)]
            sigma = self.pars["sigma_{}".format(ii)]
            total += A * np.exp(-theta2 / (2 * sigma ** 2))
        return self.pars["scale"] * total

    def to_MultiGauss2D(self, normalize=True):
        """Use this to compute containment angles and fractions.

        Note: We have to set norm = 2 * A * sigma ^ 2, because in
        MultiGauss2D norm represents the integral, and in HESS A
        represents the amplitude at 0."""
        sigmas, norms = [], []
        for ii in range(1, self.n_gauss() + 1):
            A = self.pars["A_{}".format(ii)]
            sigma = self.pars["sigma_{}".format(ii)]
            norm = self.pars["scale"] * 2 * A * sigma ** 2
            sigmas.append(sigma)
            norms.append(norm)
        m = MultiGauss2D(sigmas, norms)
        if normalize:
            m.normalize()
        return m


def multi_gauss_psf_kernel(psf_parameters, BINSZ=0.02, NEW_BINSZ=0.02, **kwargs):
    """Create multi-Gauss PSF kernel.

    The Gaussian PSF components are specified via the
    amplitude at the center and the FWHM.
    See the example for the exact format.

    Parameters
    ----------
    psf_parameters : dict
        PSF parameters
    BINSZ : float (0.02)
        Pixel size used for the given parameters in deg.
    NEW_BINSZ : float (0.02)
        New pixel size in deg. USed to change the resolution of the PSF.

    Returns
    -------
    psf_kernel : `astropy.convolution.Kernel2D`
        PSF kernel

    Examples
    --------
    >>> psf_pars = dict()
    >>> psf_pars['psf1'] = dict(ampl=1, fwhm=2.5)
    >>> psf_pars['psf2'] = dict(ampl=0.06, fwhm=11.14)
    >>> psf_pars['psf3'] = dict(ampl=0.47, fwhm=5.16)
    >>> psf_kernel = multi_gauss_psf_kernel(psf_pars, x_size=51)
    """
    psf = None
    for ii in range(1, 4):
        # Convert sigma and amplitude
        pars = psf_parameters["psf{}".format(ii)]
        sigma = gaussian_fwhm_to_sigma * pars["fwhm"] * BINSZ / NEW_BINSZ
        ampl = 2 * np.pi * sigma ** 2 * pars["ampl"]
        if psf is None:
            psf = float(ampl) * Gaussian2DKernel(sigma, **kwargs)
        else:
            psf += float(ampl) * Gaussian2DKernel(sigma, **kwargs)
    psf.normalize()
    return psf



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import absolute_import, division, print_function, unicode_literals
from collections import OrderedDict
import numpy as np
import astropy.units as u
from astropy.table import Table
from astropy.time import Time
from ..spectrum.utils import CountsPredictor
from ..stats.poisson import excess_error, excess_ul_helene
from ..utils.scripts import make_path
from ..utils.table import table_from_row_data
from ..stats.poisson import significance_on_off

__all__ = ["LightCurve", "LightCurveEstimator"]


class LightCurve(object):
    """Lightcurve container.

    The lightcurve data is stored in ``table``.

    For now we only support times stored in MJD format!

    TODO: specification of format is work in progress
    See https://github.com/open-gamma-ray-astro/gamma-astro-data-formats/pull/61

    Usage: :ref:`time-lc`

    Parameters
    ----------
    table : `~astropy.table.Table`
        Table with lightcurve data
    """

    def __init__(self, table):
        self.table = table

    def __repr__(self):
        return "{}(len={})".format(self.__class__.__name__, len(self.table))

    @property
    def time_scale(self):
        """Time scale (str).

        Taken from table "TIMESYS" header.
        Common values: "TT" or "UTC".
        Assumed default is "UTC".
        """
        return self.table.meta.get("TIMESYS", "utc")

    @property
    def time_format(self):
        """Time format (str)."""
        return "mjd"

    # @property
    # def time_ref(self):
    #     """Time reference (`~astropy.time.Time`)."""
    #     return time_ref_from_dict(self.table.meta)

    def _make_time(self, colname):
        val = self.table[colname].data
        scale = self.time_scale
        format = self.time_format
        return Time(val, scale=scale, format=format)

    @property
    def time(self):
        """Time (`~astropy.time.Time`)."""
        return self._make_time("time")

    @property
    def time_min(self):
        """Time bin start (`~astropy.time.Time`)."""
        return self._make_time("time_min")

    @property
    def time_max(self):
        """Time bin end (`~astropy.time.Time`)."""
        return self._make_time("time_max")

    @property
    def time_mid(self):
        """Time bin center (`~astropy.time.Time`)."""
        return self.time_min + 0.5 * self.time_delta

    @property
    def time_delta(self):
        """Time bin width (`~astropy.time.TimeDelta`)."""
        return self.time_max - self.time_min

    @classmethod
    def read(cls, filename, **kwargs):
        """Read from file.

        Parameters
        ----------
        filename : str
            Filename
        kwargs : dict
            Keyword arguments passed to `astropy.table.Table.read`.
        """
        filename = make_path(filename)
        table = Table.read(str(filename), **kwargs)
        return cls(table=table)

    def write(self, filename, **kwargs):
        """Write to file.

        Parameters
        ----------
        filename : str
            Filename
        kwargs : dict
            Keyword arguments passed to `astropy.table.Table.write`.
        """
        filename = make_path(filename)
        self.table.write(str(filename), **kwargs)

    def compute_fvar(self):
        r"""Calculate the fractional excess variance.

        This method accesses the the ``FLUX`` and ``FLUX_ERR`` columns
        from the lightcurve data.

        The fractional excess variance :math:`F_{var}`, an intrinsic
        variability estimator, is given by

        .. math::
            F_{var} = \sqrt{\frac{S^{2} - \bar{\sigma^{2}}}{\bar{x}^{2}}}.

        It is the excess variance after accounting for the measurement errors
        on the light curve :math:`\sigma`. :math:`S` is the variance.

        Returns
        -------
        fvar, fvar_err : `~numpy.ndarray`
            Fractional excess variance.

        References
        ----------
        .. [Vaughan2003] "On characterizing the variability properties of X-ray light
           curves from active galaxies", Vaughan et al. (2003)
           http://adsabs.harvard.edu/abs/2003MNRAS.345.1271V
        """
        flux = self.table["flux"].data.astype("float64")
        flux_err = self.table["flux_err"].data.astype("float64")

        flux_mean = np.mean(flux)
        n_points = len(flux)

        s_square = np.sum((flux - flux_mean) ** 2) / (n_points - 1)
        sig_square = np.nansum(flux_err ** 2) / n_points
        fvar = np.sqrt(np.abs(s_square - sig_square)) / flux_mean

        sigxserr_a = np.sqrt(2 / n_points) * (sig_square / flux_mean) ** 2
        sigxserr_b = np.sqrt(sig_square / n_points) * (2 * fvar / flux_mean)
        sigxserr = np.sqrt(sigxserr_a ** 2 + sigxserr_b ** 2)
        fvar_err = sigxserr / (2 * fvar)

        return fvar, fvar_err

    def compute_chisq(self):
        """Calculate the chi-square test for `LightCurve`.

        Chisquare test is a variability estimator. It computes
        deviations from the expected value here mean value

        Returns
        -------
        ChiSq, P-value : tuple of float or `~numpy.ndarray`
            Tuple of Chi-square and P-value
        """
        import scipy.stats as stats

        flux = self.table["flux"]
        yexp = np.mean(flux)
        yobs = flux.data
        chi2, pval = stats.chisquare(yobs, yexp)
        return chi2, pval

    def plot(self, ax=None, time_format="mjd", flux_unit="cm-2 s-1", **kwargs):
        """Plot flux points.

        Parameters
        ----------
        ax : `~matplotlib.axes.Axes`, optional.
            The `~matplotlib.axes.Axes` object to be drawn on.
            If None, uses the current `~matplotlib.axes.Axes`.
        time_format : {'mjd', 'iso'}, optional
            If 'iso', the x axis will contain Matplotlib dates.
            For formatting these dates see: https://matplotlib.org/gallery/ticks_and_spines/date_demo_rrule.html
        flux_unit : str, `~astropy.units.Unit`, optional
            Unit of the flux axis
        kwargs : dict
            Keyword arguments passed to :func:`matplotlib.pyplot.errorbar`

        Returns
        -------
        ax : `~matplotlib.axes.Axes`
            Axis object
        """
        import matplotlib.pyplot as plt
        from matplotlib.dates import DateFormatter

        if ax is None:
            ax = plt.gca()

        y, yerr = self._get_fluxes_and_errors(flux_unit)
        is_ul, yul = self._get_flux_uls(flux_unit)
        x, xerr = self._get_times_and_errors(time_format)

        # length of the ul arrow
        ul_arr = (
            np.nanmax(np.concatenate((y[~is_ul], yul[is_ul])))
            - np.nanmin(np.concatenate((y[~is_ul], yul[is_ul])))
        ) * 0.1

        # join fluxes and upper limits for the plot
        y[is_ul] = yul[is_ul]
        yerr[0][is_ul] = ul_arr

        # set plotting defaults and plot
        kwargs.setdefault("marker", "+")
        kwargs.setdefault("ls", "None")

        ax.errorbar(x=x, y=y, xerr=xerr, yerr=yerr, uplims=is_ul, **kwargs)
        ax.set_xlabel("Time ({})".format(time_format.upper()))
        ax.set_ylabel("Flux ({0:FITS})".format(u.Unit(flux_unit)))
        if time_format == "iso":
            ax.xaxis.set_major_formatter(DateFormatter("%Y-%m-%d %H:%M:%S"))
            plt.setp(
                ax.xaxis.get_majorticklabels(),
                rotation=30,
                ha="right",
                rotation_mode="anchor",
            )

        return ax

    def _get_fluxes_and_errors(self, unit="cm-2 s-1"):
        """Extract fluxes and corresponding errors

        Helper function for the plot method.

        Parameters
        ----------
        unit : str, `~astropy.units.Unit`, optional
            Unit of the returned flux and errors values

        Returns
        -------
        y : `numpy.ndarray`
            Flux values
        (yn, yp) : tuple of `numpy.ndarray`
            Flux error values
        """
        y = self.table["flux"].quantity.to(unit)

        if all(k in self.table.colnames for k in ["flux_errp", "flux_errn"]):
            yp = self.table["flux_errp"].quantity.to(unit)
            yn = self.table["flux_errn"].quantity.to(unit)
        elif "flux_err" in self.table.colnames:
            yp = self.table["flux_err"].quantity.to(unit)
            yn = self.table["flux_err"].quantity.to(unit)
        else:
            yp, yn = np.zeros_like(y), np.zeros_like(y)

        return y.value, (yn.value, yp.value)

    def _get_flux_uls(self, unit="cm-2 s-1"):
        """Extract flux upper limits

        Helper function for the plot method.

        Parameters
        ----------
        unit : str, `~astropy.units.Unit`, optional
            Unit of the returned flux upper limit values

        Returns
        -------
        is_ul : `numpy.ndarray`
            Is flux point is an upper limit? (boolean array)
        yul : `numpy.ndarray`
            Flux upper limit values
        """
        try:
            is_ul = self.table["is_ul"].data.astype("bool")
        except KeyError:
            is_ul = np.zeros_like(self.table["flux"]).data.astype("bool")

        if is_ul.any():
            yul = self.table["flux_ul"].quantity.to(unit)
        else:
            yul = np.zeros_like(self.table["flux"]).quantity
            yul[:] = np.nan

        return is_ul, yul.value

    def _get_times_and_errors(self, time_format="mjd"):
        """Extract times and corresponding errors

        Helper function for the plot method.

        Parameters
        ----------
        time_format : {'mjd', 'iso'}, optional
            Time format of the times. If 'iso', times and errors will be
            returned as `~datetime.datetime` and `~datetime.timedelta` objects

        Returns
        -------
        x : `~numpy.ndarray` or of `~datetime.datetime`
            Time values or `~datetime.datetime` instances if 'iso' is chosen
            as time format
        (xn, xp) : tuple of `numpy.ndarray` of `~datetime.timedelta`
            Tuple of time error values or `~datetime.timedelta` instances if
            'iso' is chosen as time format
        """
        from matplotlib.dates import num2timedelta

        try:
            x = self.time
        except KeyError:
            x = self.time_mid

        try:
            xn, xp = x - self.time_min, self.time_max - x
        except KeyError:
            xn, xp = x - x, x - x

        # convert to given time format
        if time_format == "iso":
            x = x.datetime
            # TODO: In astropy version >3.1 the TimeDelta.to_datetime() method
            # should start working, for now i will use num2timedelta.
            xn = np.asarray(num2timedelta(xn.to("d").value))
            xp = np.asarray(num2timedelta(xp.to("d").value))
        elif time_format == "mjd":
            x = x.mjd
            xn, xp = xn.to("d").value, xp.to("d").value
        else:
            raise ValueError(
                "The time format '{}' is not supported.".format(time_format)
            )

        return x, (xn, xp)


class LightCurveEstimator(object):
    """Light curve estimator.

    For a usage example see :gp-notebook:`light_curve`.

    Parameters
    ----------
    spec_extract : `~gammapy.spectrum.SpectrumExtraction`
       Contains statistics, IRF and event lists
    """

    def __init__(self, spec_extract):
        self.observations = spec_extract.observations
        self.obs_spec = spec_extract.spectrum_observations
        self.off_evt_list = self._get_off_evt_list(spec_extract)
        self.on_evt_list = self._get_on_evt_list(spec_extract)

    @staticmethod
    def _get_off_evt_list(spec_extract):
        """
        Returns list of OFF events for each observations
        """
        off_evt_list = []
        for bg in spec_extract.bkg_estimate:
            off_evt_list.append(bg.off_events)
        return off_evt_list

    @staticmethod
    def _get_on_evt_list(spec_extract):
        """
        Returns list of ON events for each observations
        """
        on_evt_list = []
        for obs in spec_extract.bkg_estimate:
            on_evt_list.append(obs.on_events)

        return on_evt_list

    @staticmethod
    def make_time_intervals_fixes(time_step, spectrum_extraction):
        """Create time intervals of fixed size.

        Parameters
        ----------
        time_step : float
            Size of the light curve bins in seconds
        spectrum_extraction : `~gammapy.spectrum.SpectrumExtraction`
            Contains statistics, IRF and event lists

        Returns
        -------
        table : `~astropy.table.Table`
            Table of time intervals

        Examples
        --------
        To extract intervals for light curve::

            intervals = list(zip(table['t_start'], table['t_stop']))
        """
        rows = []
        time_start = Time(100000, format="mjd")
        time_end = Time(0, format="mjd")
        time_step = time_step / (24 * 3600)

        for obs in spectrum_extraction.observations:
            time_events = obs.events.time
            if time_start > time_events.min():
                time_start = time_events.min()
            if time_end < time_events.max():
                time_end = time_events.max()

        time = time_start.value
        while time < time_end.value:
            time += time_step
            rows.append(
                dict(
                    t_start=Time(time - time_step, format="mjd", scale="tt"),
                    t_stop=Time(time, format="mjd", scale="tt"),
                )
            )
        return Table(rows=rows)

    def _create_and_filter_onofflists(
        self, t_index, energy_range=None, interval=None, extra=False
    ):
        """Extract on and off events list from an observation and apply energy and time filters

        Helper function for compute_flux_point and make_time_intervals_min_significance

        Parameters
        ----------
        t_index : int
            index in self of the observation to use
        energy_range : `~astropy.units.Quantity`
            True energy range to filter the events
        interval : `~astropy.time.Time`
            Time interval (2-element array)
        extra : boolean
            Define if we want spec and e_reco in output

        Returns
        -------
        on : `~gammapy.data.EventList`
            List of on events
        off : `~gammapy.data.EventList`
            List of on events
        """
        spec = self.obs_spec[t_index]
        # get ON and OFF evt list
        off = self.off_evt_list[t_index]
        on = self.on_evt_list[t_index]
        # introduce the e_reco binning here
        e_reco = spec.e_reco
        if energy_range is not None:
            emin = e_reco[e_reco.searchsorted(max(spec.lo_threshold, energy_range[0]))]
            emax = e_reco[
                e_reco.searchsorted(min(spec.hi_threshold, energy_range[1])) - 1
            ]
            # filter the event list with energy
            on = on.select_energy([emin, emax])
            on = on.select_energy(energy_range)
            off = off.select_energy([emin, emax])
            off = off.select_energy(energy_range)
        if interval is not None:
            # filter the event list with time
            tmin = interval[0]
            tmax = interval[1]
            on = on.select_time([tmin, tmax])
            off = off.select_time([tmin, tmax])
        if extra:
            return on, off, spec, e_reco
        return on, off

    @staticmethod
    def _alpha(time_holder, obs_properties, n, istart, i):
        """ Helper function for make_time_intervals_min_significance

        Parameters
        ----------
        time_holder : list of float and flag
            Contains a list of a time and a flag in 2-element arrays
        obs_properties : `~astropy.table.Table`
            Contains the dead time fraction and ratio of the on/off region
        n : int
            First observation to use
        istart : int
            index of the first event of the interval
        i : int
            index of the last event of the interval
        """

        def in_list(item, L):
            o, j = np.where(L == item)
            for index in o:
                yield index

        alpha = 0
        tmp = 0
        time = 0
        xm1 = istart
        # loop over observations
        for x in in_list("end", time_holder[istart : i + 1]):
            if tmp == 0:
                alpha += (
                    (1 - obs_properties["deadtime"][n])
                    * (
                        float(time_holder[x][0])
                        - (float(time_holder[xm1][0]) + float(time_holder[xm1 - 1][0]))
                        / 2
                    )
                    * obs_properties["A_off"][n + tmp]
                )
                time += (1 - obs_properties["deadtime"][n]) * (
                    float(time_holder[x][0])
                    - (float(time_holder[xm1][0]) + float(time_holder[xm1 - 1][0])) / 2
                )
                xm1 = x + 1
                tmp += 1
            else:
                alpha += (
                    (1 - obs_properties["deadtime"][n + tmp])
                    * (float(time_holder[x][0]) - float(time_holder[xm1][0]))
                    * obs_properties["A_off"][n + tmp]
                )
                time += (1 - obs_properties["deadtime"][n + tmp]) * (
                    float(time_holder[x][0]) - float(time_holder[xm1][0])
                )
                xm1 = x + 1
                tmp += 1
        alpha += (
            (1 - obs_properties["deadtime"][n + tmp])
            * (
                (float(time_holder[i][0]) + float(time_holder[i][0])) / 2
                - float(time_holder[xm1][0])
            )
            * obs_properties["A_off"][n + tmp]
        )
        time += (1 - obs_properties["deadtime"][n + tmp]) * (
            (float(time_holder[i][0]) + float(time_holder[i][0])) / 2
            - float(time_holder[xm1][0])
        )
        alpha = time / alpha
        return alpha

    def make_time_intervals_min_significance(
        self,
        significance,
        significance_method,
        energy_range,
        spectrum_extraction,
        separators=None,
    ):
        """
        Create time intervals such that each bin of a light curve reach a given significance

        The function work event by event to create an interval containing enough statistic and then starting a new one

        Parameters
        ----------
        significance : float
            Target significance for each light curve point
        significance_method : {'lima', 'simple'}
            Significance method (see `~gammapy.stats.significance_on_off`)
        energy_range : `~astropy.units.Quantity`
            True energy range to evaluate integrated flux (true energy)
        spectrum_extraction : `~gammapy.spectrum.SpectrumExtraction`
            Contains statistics, IRF and event lists
        separators : list of `~astropy.time.Time`
            Contains a list of time to stop the current point creation (not saved) and start a new one
            Mostly useful between observations separated by a large time gap

        Returns
        -------
        table : `~astropy.table.Table`
            Table of time intervals  and information about their content : on/off events, alpha, significance

        Examples
        --------
        extract intervals for light curve :
            intervals = list(zip(table['t_start'], table['t_stop']))
        """
        # The function create a list of time associated with identifiers called time_holder.
        # The identifiers can be 'on' for the on events, 'off' for the off events, 'start' for the start of an
        # observation, 'end for the end of an observation and 'break' for a separator.
        # The function then loop other all the elements of this list

        time_holder = []
        obs_properties = []
        n_obs = 0

        # extract the separators
        if separators is not None:
            for time in separators:
                time_holder.append([time.tt.mjd, "break"])

        # recovers the starting and ending time of each observations and useful properties
        for obs in spectrum_extraction.observations:
            time_holder.append([obs.events.observation_time_start.tt.mjd, "start"])
            time_holder.append([obs.events.observation_time_end.tt.mjd, "end"])
            obs_properties.append(
                dict(
                    deadtime=obs.observation_dead_time_fraction,
                    A_off=spectrum_extraction.bkg_estimate[n_obs].a_off,
                )
            )
            n_obs += 1
        obs_properties = Table(rows=obs_properties)

        # prepare the on and off photon list as in the flux point computation -> should be updated accordingly
        for t_index, obs in enumerate(self.observations):
            on, off = self._create_and_filter_onofflists(
                t_index=t_index, energy_range=energy_range
            )

            for time in on.time.tt.mjd:
                time_holder.append([time, "on"])
            for time in off.time.tt.mjd:
                time_holder.append([time, "off"])

        # sort all elements in the table by time
        time_holder = sorted(time_holder, key=lambda item: item[0])
        time_holder = np.asarray(time_holder)

        rows = []
        istart = 1
        i = 1
        n = 0
        while time_holder[i][0] < time_holder[-1][0]:
            i += 1
            if time_holder[i][1] == "break":
                while time_holder[i + 1][1] != "on" and time_holder[i + 1][1] != "off":
                    i += 1
                n += np.sum(time_holder[istart:i] == "end")
                istart = i
                continue
            if time_holder[i][1] != "on" and time_holder[i][1] != "off":
                continue

            # compute alpha
            alpha = self._alpha(time_holder, obs_properties, n, istart, i)

            non = np.sum(time_holder[istart : i + 1] == "on")
            noff = np.sum(time_holder[istart : i + 1] == "off")

            # check the significance
            signif = significance_on_off(non, noff, alpha, method=significance_method)
            if signif > significance:
                rows.append(
                    dict(
                        t_start=(
                            float(time_holder[istart - 1][0])
                            + float(time_holder[istart][0])
                        )
                        / 2,
                        t_stop=(float(time_holder[i][0]) + float(time_holder[i + 1][0]))
                        / 2,
                        n_on=non,
                        n_off=noff,
                        alpha=alpha,
                        significance=signif,
                    )
                )
                # start the next interval
                while (
                    time_holder[i + 1][0] < time_holder[-1][0]
                    and time_holder[i + 1][1] != "on"
                    and time_holder[i + 1][1] != "off"
                ):
                    i += 1
                n += np.sum(time_holder[istart : i + 1] == "end")
                istart = i + 1
                i = istart
        table = Table(rows=rows)
        table["t_start"] = Time(table["t_start"], format="mjd", scale="tt")
        table["t_stop"] = Time(table["t_stop"], format="mjd", scale="tt")
        return table

    def light_curve(
        self, time_intervals, spectral_model, energy_range, ul_significance=3
    ):
        """Compute light curve.

        Implementation follows what is done in:
        http://adsabs.harvard.edu/abs/2010A%26A...520A..83H.

        To be discussed: assumption that threshold energy in the
        same in reco and true energy.

        Parameters
        ----------
        time_intervals : list of `~astropy.time.Time`
            List of time intervals
        spectral_model : `~gammapy.spectrum.models.SpectralModel`
            Spectral model
        energy_range : `~astropy.units.Quantity`
            True energy range to evaluate integrated flux (true energy)
        ul_significance : float
            Upper limit confidence level significance

        Returns
        -------
        lc : `~gammapy.time.LightCurve`
            Light curve
        """
        rows = []
        for time_interval in time_intervals:
            useinterval, row = self.compute_flux_point(
                time_interval, spectral_model, energy_range, ul_significance
            )
            if useinterval:
                rows.append(row)

        table = table_from_row_data(rows)

        # TODO: check why stripping of the units here is needed
        table["time_min"] = table["time_min"].data
        table["time_max"] = table["time_max"].data
        return LightCurve(table)

    def compute_flux_point(
        self, time_interval, spectral_model, energy_range, ul_significance=3
    ):
        """Compute one flux point for one time interval.

        Parameters
        ----------
        time_interval : `~astropy.time.Time`
            Time interval (2-element array, or a tuple of Time objects)
        spectral_model : `~gammapy.spectrum.models.SpectralModel`
            Spectral model
        energy_range : `~astropy.units.Quantity`
            True energy range to evaluate integrated flux (true energy)
        ul_significance : float
            Upper limit confidence level significance

        Returns
        -------
        useinterval : bool
            Is True if the time_interval produce a valid flux point
        measurements : dict
            Dictionary with flux point measurement in the time interval
        """
        tmin, tmax = time_interval[0], time_interval[1]
        livetime = 0
        alpha_mean = 0.0
        alpha_mean_backup = 0.0
        measured_excess = 0
        predicted_excess = 0
        n_on = 0
        n_off = 0
        useinterval = False

        # Loop on observations
        for t_index, obs in enumerate(self.observations):

            # discard observations not matching the time interval
            obs_start = obs.events.observation_time_start
            obs_stop = obs.events.observation_time_end
            if (tmin < obs_start and tmax < obs_start) or (tmin > obs_stop):
                continue
            useinterval = True
            # get ON and OFF evt list
            on, off, spec, e_reco = self._create_and_filter_onofflists(
                t_index=t_index,
                energy_range=energy_range,
                interval=[tmin, tmax],
                extra=True,
            )
            n_on_obs = len(on.table)
            n_off_obs = len(off.table)

            # compute effective livetime (for the interval)
            if tmin >= obs_start and tmax <= obs_stop:
                # interval included in obs
                livetime_to_add = (tmax - tmin).to("s")
            elif tmin >= obs_start and tmax >= obs_stop:
                # interval min above tstart from obs
                livetime_to_add = (obs_stop - tmin).to("s")
            elif tmin <= obs_start and tmax <= obs_stop:
                # interval min below tstart from obs
                livetime_to_add = (tmax - obs_start).to("s")
            elif tmin <= obs_start and tmax >= obs_stop:
                # obs included in interval
                livetime_to_add = (obs_stop - obs_start).to("s")
            else:
                livetime_to_add = u.Quantity(0, "s")

            # Take into account dead time
            livetime_to_add *= 1.0 - obs.observation_dead_time_fraction

            # Compute excess
            obs_measured_excess = n_on_obs - spec.alpha * n_off_obs

            # Compute the expected excess in the range given by the user
            # but must respect the energy threshold of the observation
            # (to match the energy range of the measured excess)
            # We use the effective livetime and the right energy threshold
            e_idx = np.where(
                np.logical_and.reduce(
                    (
                        e_reco >= spec.lo_threshold,  # threshold
                        e_reco <= spec.hi_threshold,  # threshold
                        e_reco >= energy_range[0],  # user
                        e_reco <= energy_range[-1],
                    )  # user
                )
            )[0]
            counts_predictor = CountsPredictor(
                livetime=livetime_to_add,
                aeff=spec.aeff,
                edisp=spec.edisp,
                model=spectral_model,
            )
            counts_predictor.run()
            counts_predicted_excess = counts_predictor.npred.data.data[e_idx[:-1]]

            obs_predicted_excess = np.sum(counts_predicted_excess)

            # compute effective normalisation between ON/OFF (for the interval)
            livetime += livetime_to_add
            alpha_mean += spec.alpha * n_off_obs
            alpha_mean_backup += spec.alpha * livetime_to_add
            measured_excess += obs_measured_excess
            predicted_excess += obs_predicted_excess
            n_on += n_on_obs
            n_off += n_off_obs

        # Fill time interval information
        if useinterval:
            int_flux = spectral_model.integral(energy_range[0], energy_range[1])

            if n_off > 0.0:
                alpha_mean /= n_off
            if livetime > 0.0:
                alpha_mean_backup /= livetime
            if alpha_mean == 0.0:  # use backup if necessary
                alpha_mean = alpha_mean_backup

            flux = measured_excess / predicted_excess.value
            flux *= int_flux

            # Gaussian errors, TODO: should be improved
            flux_err = int_flux / predicted_excess.value
            delta_excess = excess_error(n_on=n_on, n_off=n_off, alpha=alpha_mean)
            flux_err *= delta_excess

            sigma = significance_on_off(
                n_on=n_on, n_off=n_off, alpha=alpha_mean, method="lima"
            )
            is_ul = sigma <= 3
            flux_ul = excess_ul_helene(
                n_on - alpha_mean * n_off, delta_excess, ul_significance
            )
            flux_ul *= int_flux / predicted_excess.value
        else:
            flux = u.Quantity(0, "cm-2 s-1")
            flux_err = u.Quantity(0, "cm-2 s-1")
            flux_ul = u.Quantity(-1, "cm-2 s-1")

        # Store measurements in a dict and return that
        return (
            useinterval,
            OrderedDict(
                [
                    # TODO: check why time is stored without units
                    ("time_min", Time(tmin, format="mjd").value),
                    ("time_max", Time(tmax, format="mjd").value),
                    ("flux", flux),
                    ("flux_err", flux_err),
                    ("flux_ul", flux_ul),
                    ("is_ul", is_ul),
                    ("livetime", livetime),
                    ("alpha", alpha_mean),
                    ("n_on", n_on),
                    ("n_off", n_off),
                    ("measured_excess", measured_excess),
                    ("expected_excess", predicted_excess),
                ]
            ),
        )



"""Fitting framework and backends."""
# Core
from .parameter import *
from .model import *
from .likelihood import *
from .fit import *

# Backends
from .scipy import *
from .iminuit import *
from .sherpa import *



# Licensed under a 3-clause BSD style license - see LICENSE.rst
from __future__ import absolute_import, division, print_function, unicode_literals
import copy
import numpy as np
from astropy.units import Quantity
from ..extern.six.moves import UserList  # pylint:disable=import-error
from ..extern.pathlib import Path
from ..utils.scripts import make_path
from ..utils.energy import EnergyBounds
from ..utils.table import table_from_row_data
from ..data import ObservationStats
from ..irf import EffectiveAreaTable, EnergyDispersion, IRFStacker
from .core import CountsSpectrum, PHACountsSpectrum, PHACountsSpectrumList
from .utils import CountsPredictor

__all__ = [
    "SpectrumStats",
    "SpectrumObservation",
    "SpectrumObservationList",
    "SpectrumObservationStacker",
]


class SpectrumStats(ObservationStats):
    """Spectrum stats.

    Extends `~gammapy.data.ObservationStats` with spectrum
    specific information (energy bin info at the moment).
    """

    def __init__(self, **kwargs):
        self.energy_min = kwargs.pop("energy_min", Quantity(0, "TeV"))
        self.energy_max = kwargs.pop("energy_max", Quantity(0, "TeV"))
        super(SpectrumStats, self).__init__(**kwargs)

    def __str__(self):
        ss = super(SpectrumStats, self).__str__()
        ss += "energy range: {:.2f} - {:.2f}".format(self.energy_min, self.energy_max)
        return ss

    def to_dict(self):
        """TODO: document"""
        data = super(SpectrumStats, self).to_dict()
        data["energy_min"] = self.energy_min
        data["energy_max"] = self.energy_max
        return data


class SpectrumObservation(object):
    """1D spectral analysis storage class.

    This container holds the ingredients for 1D region based spectral analysis.

    Meta data is stored in the ``on_vector`` attribute.
    This reflects the OGIP convention.

    Parameters
    ----------
    on_vector : `~gammapy.spectrum.PHACountsSpectrum`
        On vector
    aeff : `~gammapy.irf.EffectiveAreaTable`
        Effective Area
    off_vector : `~gammapy.spectrum.PHACountsSpectrum`, optional
        Off vector
    edisp : `~gammapy.irf.EnergyDispersion`, optional
        Energy dispersion matrix

    Examples
    --------
    ::

        from gammapy.spectrum import SpectrumObservation
        filename = '$GAMMAPY_DATA/joint-crab/spectra/hess/pha_obs23523.fits'
        obs = SpectrumObservation.read(filename)
        print(obs)
    """

    def __init__(self, on_vector, aeff=None, off_vector=None, edisp=None):
        self.on_vector = on_vector
        self.aeff = aeff
        self.off_vector = off_vector
        self.edisp = edisp

    def __str__(self):
        ss = self.total_stats_safe_range.__str__()
        return ss

    @property
    def obs_id(self):
        """Unique identifier"""
        return self.on_vector.obs_id

    @obs_id.setter
    def obs_id(self, obs_id):
        self.on_vector.obs_id = obs_id
        if self.off_vector is not None:
            self.off_vector.obs_id = obs_id

    @property
    def meta(self):
        """Meta information"""
        return self.on_vector.meta

    @property
    def livetime(self):
        """Dead-time corrected observation time"""
        return self.on_vector.livetime

    @property
    def alpha(self):
        """Exposure ratio between signal and background regions"""
        return self.on_vector.backscal / self.off_vector.backscal

    @property
    def e_reco(self):
        """Reconstruced energy bounds array."""
        return EnergyBounds(self.on_vector.energy.bins)

    @property
    def e_true(self):
        """True energy bounds array."""
        return EnergyBounds(self.aeff.energy.bins)

    @property
    def nbins(self):
        """Number of reconstruced energy bins"""
        return self.on_vector.energy.nbins

    @property
    def lo_threshold(self):
        """Low energy threshold"""
        return self.on_vector.lo_threshold

    @lo_threshold.setter
    def lo_threshold(self, threshold):
        self.on_vector.lo_threshold = threshold
        if self.off_vector is not None:
            self.off_vector.lo_threshold = threshold

    @property
    def hi_threshold(self):
        """High energy threshold"""
        return self.on_vector.hi_threshold

    @hi_threshold.setter
    def hi_threshold(self, threshold):
        self.on_vector.hi_threshold = threshold
        if self.off_vector is not None:
            self.off_vector.hi_threshold = threshold

    def reset_thresholds(self):
        """Reset energy thresholds (i.e. declare all energy bins valid)"""
        self.on_vector.reset_thresholds()
        if self.off_vector is not None:
            self.off_vector.reset_thresholds()

    def compute_energy_threshold(
        self, method_lo="none", method_hi="none", reset=False, **kwargs
    ):
        """Compute and set the safe energy threshold.

        Set the high and low energy threshold for each observation based on a
        chosen method.

        Available methods for setting the low energy threshold:

        * area_max : Set energy threshold at x percent of the maximum effective
          area (x given as kwargs['area_percent_lo'])

        * energy_bias : Set energy threshold at energy where the energy bias
          exceeds a value of x percent (given as kwargs['bias_percent_lo'])

        * none : Do not apply a lower threshold

        Available methods for setting the high energy threshold:

        * area_max : Set energy threshold at x percent of the maximum effective
          area (x given as kwargs['area_percent_hi'])

        * energy_bias : Set energy threshold at energy where the energy bias
          exceeds a value of x percent (given as kwargs['bias_percent_hi'])

        * none : Do not apply a higher energy threshold

        Parameters
        ----------
        method_lo : {'area_max', 'energy_bias', 'none'}
            Method for defining the low energy threshold

        method_hi : {'area_max', 'energy_bias', 'none'}
            Method for defining the high energy threshold

        reset : bool
            Reset existing energy thresholds before setting the new ones
            (default is `False`)
        """

        if reset:
            self.reset_thresholds()

        # It is important to update the low and high threshold for ON and OFF
        # vector, otherwise Sherpa will not understand the files

        # Low threshold
        if method_lo == "area_max":
            aeff_thres = kwargs["area_percent_lo"] / 100 * self.aeff.max_area
            thres_lo = self.aeff.find_energy(aeff_thres)
        elif method_lo == "energy_bias":
            thres_lo = self._find_bias_energy(kwargs["bias_percent_lo"] / 100)
        elif method_lo == "none":
            thres_lo = self.e_true[0]
        else:
            raise ValueError("Undefine method for low threshold: {}".format(method_lo))

        self.on_vector.lo_threshold = thres_lo
        if self.off_vector is not None:
            self.off_vector.lo_threshold = thres_lo

        # High threshold
        if method_hi == "area_max":
            aeff_thres = kwargs["area_percent_hi"] / 100 * self.aeff.max_area
            thres_hi = self.aeff.find_energy(aeff_thres, reverse=True)
        elif method_hi == "energy_bias":
            thres_hi = self._find_bias_energy(
                kwargs["bias_percent_hi"] / 100, reverse=True
            )
        elif method_hi == "none":
            thres_hi = self.e_true[-1]
        else:
            raise ValueError(
                "Undefined method for high threshold: {}".format(method_hi)
            )

        self.on_vector.hi_threshold = thres_hi
        if self.off_vector is not None:
            self.off_vector.hi_threshold = thres_hi

    def _find_bias_energy(self, bias_value, reverse=False):
        """Helper function to interpolate between bias values to retrieve an energy"""
        e = self.e_true.log_centers
        bias = np.abs(self.edisp.get_bias(e))
        with np.errstate(invalid="ignore"):
            valid = np.where(bias <= bias_value)[0]
        idx = valid[0]
        if reverse:
            idx = valid[-1]
        if not reverse:
            if idx == 0:
                energy = self.e_true[idx].value
            else:
                energy = np.interp(
                    bias_value, (bias[[idx - 1, idx]].value), (e[[idx - 1, idx]].value)
                )
        else:
            if idx == e.size - 1:
                energy = self.e_true[idx + 1].value
            else:
                energy = np.interp(
                    bias_value, (bias[[idx, idx + 1]].value), (e[[idx, idx + 1]].value)
                )
        return energy * self.e_true.unit

    @property
    def background_vector(self):
        """Background `~gammapy.spectrum.CountsSpectrum`.

        bkg = alpha * n_off

        If alpha is a function of energy this will differ from
        self.on_vector * self.total_stats.alpha because the latter returns an
        average value for alpha.
        """
        energy = self.off_vector.energy
        data = self.off_vector.data.data * self.alpha
        return CountsSpectrum(data=data, energy_lo=energy.lo, energy_hi=energy.hi)

    @property
    def excess_vector(self):
        """Excess `~gammapy.spectrum.CountsSpectrum`.

        excess = n_on = alpha * n_off
        """
        energy = self.off_vector.energy
        data = self.on_vector.data.data - self.background_vector.data.data
        return CountsSpectrum(data=data, energy_lo=energy.lo, energy_hi=energy.hi)

    @property
    def total_stats(self):
        """Return total `~gammapy.spectrum.SpectrumStats`
        """
        return self.stats_in_range(0, self.nbins - 1)

    @property
    def total_stats_safe_range(self):
        """Return total `~gammapy.spectrum.SpectrumStats` within the tresholds
        """
        safe_bins = self.on_vector.bins_in_safe_range
        return self.stats_in_range(safe_bins[0], safe_bins[-1])

    def stats_in_range(self, bin_min, bin_max):
        """Compute stats for a range of energy bins.

        Parameters
        ----------
        bin_min, bin_max: int
            Bins to include

        Returns
        -------
        stats : `~gammapy.spectrum.SpectrumStats`
            Stacked stats
        """
        idx = np.arange(bin_min, bin_max + 1)
        stats_list = [self.stats(ii) for ii in idx]
        stacked_stats = SpectrumStats.stack(stats_list)
        stacked_stats.livetime = self.livetime
        stacked_stats.gamma_rate = stacked_stats.excess / stacked_stats.livetime
        stacked_stats.obs_id = self.obs_id
        stacked_stats.energy_min = self.e_reco[bin_min]
        stacked_stats.energy_max = self.e_reco[bin_max + 1]
        return stacked_stats

    def stats(self, idx):
        """Compute stats for one energy bin.

        Parameters
        ----------
        idx : int
            Energy bin index

        Returns
        -------
        stats : `~gammapy.spectrum.SpectrumStats`
            Stats
        """
        if self.off_vector is not None:
            n_off = int(self.off_vector.data.data.value[idx])
            a_off = self.off_vector._backscal_array[idx]
        else:
            n_off = 0
            a_off = 1  # avoid zero division error

        return SpectrumStats(
            energy_min=self.e_reco[idx],
            energy_max=self.e_reco[idx + 1],
            n_on=int(self.on_vector.data.data.value[idx]),
            n_off=n_off,
            a_on=self.on_vector._backscal_array[idx],
            a_off=a_off,
            obs_id=self.obs_id,
            livetime=self.livetime,
        )

    def stats_table(self):
        """Per-bin stats as a table.

        Returns
        -------
        table : `~astropy.table.Table`
            Table with stats for one energy bin in one row.
        """
        rows = [self.stats(idx).to_dict() for idx in range(len(self.e_reco) - 1)]
        return table_from_row_data(rows=rows)

    def predicted_counts(self, model):
        """Calculated number of predicted counts given a model.

        Parameters
        ----------
        model : `~gammapy.spectrum.models.SpectralModel`
            Spectral model

        Returns
        -------
        npred : `~gammapy.spectrum.CountsSpectrum`
            Predicted counts
        """
        predictor = CountsPredictor(
            model=model, edisp=self.edisp, aeff=self.aeff, livetime=self.livetime
        )
        predictor.run()
        return predictor.npred

    @classmethod
    def read(cls, filename):
        """Read `~gammapy.spectrum.SpectrumObservation` from OGIP files.

        BKG file, ARF, and RMF must be set in the PHA header and be present in
        the same folder.

        Parameters
        ----------
        filename : str
            OGIP PHA file to read
        """
        filename = make_path(filename)
        dirname = filename.parent
        on_vector = PHACountsSpectrum.read(filename)
        rmf, arf, bkg = on_vector.rmffile, on_vector.arffile, on_vector.bkgfile

        try:
            energy_dispersion = EnergyDispersion.read(str(dirname / rmf))
        except IOError:
            # TODO : Add logger and echo warning
            energy_dispersion = None

        try:
            off_vector = PHACountsSpectrum.read(str(dirname / bkg))
        except IOError:
            # TODO : Add logger and echo warning
            off_vector = None

        effective_area = EffectiveAreaTable.read(str(dirname / arf))

        return cls(
            on_vector=on_vector,
            aeff=effective_area,
            off_vector=off_vector,
            edisp=energy_dispersion,
        )

    def write(self, outdir=None, use_sherpa=False, overwrite=False):
        """Write OGIP files.

        If you want to use the written files with Sherpa you have to set the
        ``use_sherpa`` flag. Then all files will be written in units 'keV' and
        'cm2'.

        Parameters
        ----------
        outdir : `~gammapy.extern.pathlib.Path`
            output directory, default: pwd
        use_sherpa : bool, optional
            Write Sherpa compliant files, default: False
        overwrite : bool
            Overwrite existing files?
        """
        outdir = Path.cwd() if outdir is None else Path(outdir)
        outdir.mkdir(exist_ok=True, parents=True)

        phafile = self.on_vector.phafile
        bkgfile = self.on_vector.bkgfile
        arffile = self.on_vector.arffile
        rmffile = self.on_vector.rmffile

        # Write in keV and cm2 for sherpa
        if use_sherpa:
            # TODO: Change this implementation.
            # write should not change the object
            # put this code in a separate method that makes a copy with the changes.
            # then call `.write` on that here, or remove the option and let the user do it.
            self.on_vector.energy.lo = self.on_vector.energy.lo.to("keV")
            self.on_vector.energy.hi = self.on_vector.energy.hi.to("keV")
            self.aeff.energy.lo = self.aeff.energy.lo.to("keV")
            self.aeff.energy.hi = self.aeff.energy.hi.to("keV")
            self.aeff.data.data = self.aeff.data.data.to("cm2")
            if self.off_vector is not None:
                self.off_vector.energy.lo = self.off_vector.energy.lo.to("keV")
                self.off_vector.energy.hi = self.off_vector.energy.hi.to("keV")
            if self.edisp is not None:
                self.edisp.e_reco.lo = self.edisp.e_reco.lo.to("keV")
                self.edisp.e_reco.hi = self.edisp.e_reco.hi.to("keV")
                self.edisp.e_true.lo = self.edisp.e_true.lo.to("keV")
                self.edisp.e_true.hi = self.edisp.e_true.hi.to("keV")
                # Set data to itself to trigger reset of the interpolator
                # TODO: Make NDData notice change of axis
                self.edisp.data.data = self.edisp.data.data

        self.on_vector.write(outdir / phafile, overwrite=overwrite)
        self.aeff.write(outdir / arffile, overwrite=overwrite)
        if self.off_vector is not None:
            self.off_vector.write(outdir / bkgfile, overwrite=overwrite)
        if self.edisp is not None:
            self.edisp.write(str(outdir / rmffile), overwrite=overwrite)

    def peek(self, figsize=(10, 10)):
        """Quick-look summary plots."""
        import matplotlib.pyplot as plt

        fig, ((ax1, ax2), (ax3, ax4)) = plt.subplots(nrows=2, ncols=2, figsize=figsize)

        ax1.set_title("Counts")
        energy_unit = "TeV"
        if self.off_vector is not None:
            self.background_vector.plot_hist(
                ax=ax1, label="alpha * n_off", color="darkblue", energy_unit=energy_unit
            )
        self.on_vector.plot_hist(
            ax=ax1,
            label="n_on",
            color="darkred",
            energy_unit=energy_unit,
            show_energy=(self.hi_threshold, self.lo_threshold),
        )
        ax1.set_xlim(
            0.7 * self.lo_threshold.to_value(energy_unit),
            1.3 * self.hi_threshold.to_value(energy_unit),
        )
        ax1.legend(numpoints=1)

        ax2.set_title("Effective Area")
        e_unit = self.aeff.energy.unit
        self.aeff.plot(ax=ax2, show_energy=(self.hi_threshold, self.lo_threshold))
        ax2.set_xlim(
            0.7 * self.lo_threshold.to_value(e_unit),
            1.3 * self.hi_threshold.to_value(e_unit),
        )

        ax3.axis("off")
        if self.off_vector is not None:
            ax3.text(0, 0.2, "{}".format(self.total_stats_safe_range), fontsize=12)

        ax4.set_title("Energy Dispersion")
        if self.edisp is not None:
            self.edisp.plot_matrix(ax=ax4)

        # TODO: optimize layout
        plt.subplots_adjust(wspace=0.3)

    def to_sherpa(self):
        """Convert to `~sherpa.astro.data.DataPHA`.

        Associated background vectors and IRFs are also translated to sherpa
        objects and appended to the PHA instance.
        """
        pha = self.on_vector.to_sherpa(name="pha_obs{}".format(self.obs_id))
        if self.aeff is not None:
            arf = self.aeff.to_sherpa(name="arf_obs{}".format(self.obs_id))
        else:
            arf = None
        if self.edisp is not None:
            rmf = self.edisp.to_sherpa(name="rmf_obs{}".format(self.obs_id))
        else:
            rmf = None

        pha.set_response(arf, rmf)

        if self.off_vector is not None:
            bkg = self.off_vector.to_sherpa(name="bkg_obs{}".format(self.obs_id))
            bkg.set_response(arf, rmf)
            pha.set_background(bkg, 1)

        # see https://github.com/sherpa/sherpa/blob/36c1f9dabb3350b64d6f54ab627f15c862ee4280/sherpa/astro/data.py#L1400
        pha._set_initial_quantity()
        return pha

    def copy(self):
        """A deep copy."""
        return copy.deepcopy(self)


class SpectrumObservationList(UserList):
    """List of `~gammapy.spectrum.SpectrumObservation` objects."""

    def __str__(self):
        ss = self.__class__.__name__
        ss += "\nNumber of observations: {}".format(len(self))
        # ss += '\n{}'.format(self.obs_id)
        return ss

    def obs(self, obs_id):
        """Return one observation.

        Parameters
        ----------
        obs_id : int
            Identifier
        """
        obs_id_list = [o.obs_id for o in self]
        idx = obs_id_list.index(obs_id)
        return self[idx]

    @property
    def obs_id(self):
        """List of observations ids"""
        return [o.obs_id for o in self]

    @property
    def total_livetime(self):
        """Summed livetime"""
        livetimes = [o.livetime.to_value("s") for o in self]
        return Quantity(np.sum(livetimes), "s")

    @property
    def on_vector_list(self):
        """On `~gammapy.spectrum.PHACountsSpectrumList`"""
        return PHACountsSpectrumList([o.on_vector for o in self])

    @property
    def off_vector_list(self):
        """Off `~gammapy.spectrum.PHACountsSpectrumList`"""
        return PHACountsSpectrumList([o.off_vector for o in self])

    def stack(self):
        """Return stacked `~gammapy.spectrum.SpectrumObservation`"""
        stacker = SpectrumObservationStacker(obs_list=self)
        stacker.run()
        return stacker.stacked_obs

    def safe_range(self, method="inclusive"):
        """Safe energy range

        This is the energy range in with any / all observations have their safe
        threshold

        Parameters
        ----------
        method : str, {'inclusive', 'exclusive'}
            Maximum or minimum range
        """
        unit = "TeV"
        lo = [obs.lo_threshold.to_value(unit) for obs in self]
        hi = [obs.hi_threshold.to_value(unit) for obs in self]

        if method == "inclusive":
            return Quantity([min(lo), max(hi)], unit)
        elif method == "exclusive":
            return Quantity([max(lo), min(hi)], unit)
        else:
            raise ValueError("Invalid method: {}".format(method))

    def write(self, outdir=None, pha_typeII=False, **kwargs):
        """Create OGIP files

        Each observation will be written as seperate set of FITS files by
        default. If the option ``pha_typeII`` is enabled all on and off counts
        spectra will be collected into one
        `~gammapy.spectrum.PHACountsSpectrumList` and written to one FITS file.
        All datasets will be associated to the same response files.
        see
        https://heasarc.gsfc.nasa.gov/docs/heasarc/ofwg/docs/spectra/ogip_92_007/node8.html

        TODO: File written with the ``pha_typeII`` option are not read
        properly with sherpa. This could be a sherpa issue. Investigate and
        file issue.

        Parameters
        ----------
        outdir : str, `~gammapy.extern.pathlib.Path`, optional
            Output directory, default: pwd
        pha_typeII : bool, default: False
            Collect PHA datasets into one file
        """
        outdir = make_path(outdir)
        outdir.mkdir(exist_ok=True, parents=True)
        if not pha_typeII:
            for obs in self:
                obs.write(outdir=outdir, **kwargs)
        else:
            onlist = self.on_vector_list
            onlist.write(outdir / "pha2.fits", **kwargs)
            offlist = self.off_vector_list
            # This filename is hardcoded since it is a column in the on list
            offlist.write(outdir / "bkg.fits", **kwargs)
            arf_file = onlist.to_table().meta["ancrfile"]
            rmf_file = onlist.to_table().meta["respfile"]
            self[0].aeff.write(outdir / arf_file, **kwargs)
            self[0].edisp.write(outdir / rmf_file, **kwargs)

    @classmethod
    def read(cls, directory, pha_typeII=False):
        """Read multiple observations

        This methods reads all PHA files contained in a given directory. Enable
        ``pha_typeII`` to read a PHA type II file.

        see
        https://heasarc.gsfc.nasa.gov/docs/heasarc/ofwg/docs/spectra/ogip_92_007/node8.html

        TODO: Replace with more sophisticated file managment system

        Parameters
        ----------
        directory : `~gammapy.extern.pathlib.Path`
            Directory holding the observations
        pha_typeII : bool, default: False
            Read PHA typeII file
        """
        obs_list = cls()
        directory = make_path(directory)

        if not pha_typeII:
            # glob default order depends on OS, so we call sorted() explicitely to
            # get reproducable results
            filelist = sorted(directory.glob("pha*.fits"))
            for phafile in filelist:
                obs = SpectrumObservation.read(phafile)
                obs_list.append(obs)
        else:
            # NOTE: filenames for type II PHA files are hardcoded
            on_vectors = PHACountsSpectrumList.read(directory / "pha2.fits")
            off_vectors = PHACountsSpectrumList.read(directory / "bkg.fits")
            aeff = EffectiveAreaTable.read(directory / "arf.fits")
            edisp = EnergyDispersion.read(directory / "rmf.fits")

            for on, off in zip(on_vectors, off_vectors):
                obs = SpectrumObservation(
                    on_vector=on, off_vector=off, aeff=aeff, edisp=edisp
                )
                obs_list.append(obs)

        return obs_list

    def peek(self):
        """Quickly look at observations

        Uses IPython widgets.
        TODO: Change to bokeh
        """
        from ipywidgets import interact

        max_ = len(self) - 1

        def show_obs(idx):
            self[idx].peek()

        return interact(show_obs, idx=(0, max_, 1))


class SpectrumObservationStacker(object):
    r"""Stack observations in a `~gammapy.spectrum.SpectrumObservationList`.

    The stacking of :math:`j` observations is implemented as follows.
    :math:`k` and :math:`l` denote a bin in reconstructed and true energy,
    respectively.

    .. math::

        \epsilon_{jk} =\left\{\begin{array}{cl} 1, & \mbox{if
            bin k is inside the energy thresholds}\\ 0, & \mbox{otherwise} \end{array}\right.

        \overline{\mathrm{n_{on}}}_k = \sum_{j} \mathrm{n_{on}}_{jk} \cdot
            \epsilon_{jk}

        \overline{\mathrm{n_{off}}}_k = \sum_{j} \mathrm{n_{off}}_{jk} \cdot
            \epsilon_{jk}

        \overline{\alpha}_k =
        \frac{\overline{{b_{on}}}_k}{\overline{{b_{off}}}_k}

        \overline{{b}_{on}}_k = 1

        \overline{{b}_{off}}_k = \frac{1}{\sum_{j}\alpha_{jk} \cdot
            \mathrm{n_{off}}_{jk} \cdot \epsilon_{jk}} \cdot \overline{\mathrm {n_{off}}}

    Please refer to the `~gammapy.irf.IRFStacker` for the description
    of how the IRFs are stacked.

    Parameters
    ----------
    obs_list : `~gammapy.spectrum.SpectrumObservationList`
        Observations to stack

    Examples
    --------
    >>> from gammapy.spectrum import SpectrumObservationList, SpectrumObservationStacker
    >>> obs_list = SpectrumObservationList.read('$GAMMAPY_DATA/joint-crab/spectra/hess')
    >>> obs_stacker = SpectrumObservationStacker(obs_list)
    >>> obs_stacker.run()
    >>> print(obs_stacker.stacked_obs)
    *** Observation summary report ***
    Observation Id: [23523-23592]
    Livetime: 0.879 h
    On events: 279
    Off events: 108
    Alpha: 0.037
    Bkg events in On region: 3.96
    Excess: 275.04
    Excess / Background: 69.40
    Gamma rate: 0.14 1 / min
    Bkg rate: 0.00 1 / min
    Sigma: 37.60
    energy range: 681292069.06 keV - 87992254356.91 keV
    """

    def __init__(self, obs_list):
        self.obs_list = SpectrumObservationList(obs_list)
        self.stacked_on_vector = None
        self.stacked_off_vector = None
        self.stacked_aeff = None
        self.stacked_edisp = None
        self.stacked_bkscal_on = None
        self.stacked_bkscal_off = None
        self.stacked_obs = None

    def __str__(self):
        ss = self.__class__.__name__
        ss += "\n{}".format(self.obs_list)
        return ss

    def run(self):
        """Run all steps in the correct order."""
        self.stack_counts_vectors()
        self.stack_aeff()
        self.stack_edisp()
        self.stack_obs()

    def stack_counts_vectors(self):
        """Stack on and off vectors."""
        self.stack_on_vector()
        self.stack_off_vector()
        self.stack_backscal()
        self.setup_counts_vectors()

    def stack_on_vector(self):
        """Stack the on count vector."""
        on_vector_list = [o.on_vector for o in self.obs_list]
        self.stacked_on_vector = self.stack_counts_spectrum(on_vector_list)

    def stack_off_vector(self):
        """Stack the off count vector."""
        off_vector_list = [o.off_vector for o in self.obs_list]
        self.stacked_off_vector = self.stack_counts_spectrum(off_vector_list)

    @staticmethod
    def stack_counts_spectrum(counts_spectrum_list):
        """Stack `~gammapy.spectrum.PHACountsSpectrum`.

        * Bins outside the safe energy range are set to 0
        * Attributes are set to None.
        * The quality vector of the observations are combined with a logical or,
          such that the low (high) threshold of the stacked obs is the minimum
          low (maximum high) threshold of the observation list to be stacked.
        """
        template = counts_spectrum_list[0].copy()
        energy = template.energy
        stacked_data = np.zeros(energy.nbins)
        stacked_quality = np.ones(energy.nbins)
        for spec in counts_spectrum_list:
            stacked_data += spec.counts_in_safe_range.value
            temp = np.logical_and(stacked_quality, spec.quality)
            stacked_quality = np.array(temp, dtype=int)

        return PHACountsSpectrum(
            data=stacked_data,
            energy_lo=energy.lo,
            energy_hi=energy.hi,
            quality=stacked_quality,
        )

    def stack_backscal(self):
        """Stack ``backscal`` for on and off vector."""
        nbins = self.obs_list[0].e_reco.nbins
        bkscal_on = np.ones(nbins)
        bkscal_off = np.zeros(nbins)

        alpha_sum = 0.0

        for obs in self.obs_list:
            bkscal_on_data = obs.on_vector._backscal_array.copy()
            bkscal_off_data = obs.off_vector._backscal_array.copy()
            bkscal_off += (
                bkscal_on_data / bkscal_off_data
            ) * obs.off_vector.counts_in_safe_range.value
            alpha_sum += (obs.alpha * obs.off_vector.counts_in_safe_range).sum()

        with np.errstate(divide="ignore", invalid="ignore"):
            stacked_bkscal_off = self.stacked_off_vector.data.data.value / bkscal_off
            alpha_average = (
                alpha_sum / self.stacked_off_vector.counts_in_safe_range.sum()
            )

        # there should be no nan values in backscal_on or backscal_off
        # this leads to problems when fitting the data
        # use 1 for backscale of on_vector and 1 / alpha_average for backscale of off_vector
        alpha_correction = 1
        idx = np.where(self.stacked_off_vector.data.data == 0)[0]
        bkscal_on[idx] = alpha_correction
        # For the bins where the stacked OFF counts equal 0, the alpha value is performed by weighting on the total
        # OFF counts of each run
        stacked_bkscal_off[idx] = alpha_correction / alpha_average

        self.stacked_bkscal_on = bkscal_on
        self.stacked_bkscal_off = stacked_bkscal_off

    def setup_counts_vectors(self):
        """Add correct attributes to stacked counts vectors."""
        total_livetime = self.obs_list.total_livetime
        self.stacked_on_vector.livetime = total_livetime
        self.stacked_off_vector.livetime = total_livetime
        self.stacked_on_vector.backscal = self.stacked_bkscal_on
        self.stacked_off_vector.backscal = self.stacked_bkscal_off
        self.stacked_on_vector.obs_id = self.obs_list.obs_id
        self.stacked_off_vector.obs_id = self.obs_list.obs_id

    def stack_aeff(self):
        """Stack effective areas (weighted by livetime).

        Calls `gammapy.irf.IRFStacker.stack_aeff`.
        """
        irf_stacker = IRFStacker(
            list_aeff=[obs.aeff for obs in self.obs_list],
            list_livetime=[obs.livetime for obs in self.obs_list],
        )
        irf_stacker.stack_aeff()
        self.stacked_aeff = irf_stacker.stacked_aeff

    def stack_edisp(self):
        """Stack energy dispersion (weighted by exposure).

        Calls `~gammapy.irf.IRFStacker.stack_edisp`
        """
        irf_stacker = IRFStacker(
            list_aeff=[obs.aeff for obs in self.obs_list],
            list_livetime=[obs.livetime for obs in self.obs_list],
            list_edisp=[obs.edisp for obs in self.obs_list],
            list_low_threshold=[obs.lo_threshold for obs in self.obs_list],
            list_high_threshold=[obs.hi_threshold for obs in self.obs_list],
        )
        irf_stacker.stack_edisp()
        self.stacked_edisp = irf_stacker.stacked_edisp

    def stack_obs(self):
        """Create stacked `~gammapy.spectrum.SpectrumObservation`"""
        self.stacked_obs = SpectrumObservation(
            on_vector=self.stacked_on_vector,
            off_vector=self.stacked_off_vector,
            aeff=self.stacked_aeff,
            edisp=self.stacked_edisp,
        )
